;**************************************************************************
;*
;* FILE  monitor.asm
;* Copyright (c) 1995, 2005, 2006, 2007, 2008 Daniel Kahlin <daniel@kahlin.net>
;* Copyright (c) 1986 Per Hkan Sundell
;* Copyright (c) 1983 Human Engineered Software (H.E.S.)
;* Written by Daniel Kahlin <daniel@kahlin.net>
;* $Id: monitor.asm,v 1.83 2008-02-17 16:21:58 tlr Exp $
;*
;* DESCRIPTION
;*   based on TLRMON by Daniel Kahlin (1995)
;*   based on C.C.S. MON by Per H. Sundell (1986).
;*   based on HESMON 2.0 by Terry Peterson (1983).
;*
;* LAYOUT
;*   +----------------------------------+
;*   | Seg #0, buffers $0000-$0c00      |
;*   | $0000   (could be backed up DMA) |
;*   +----------------------------------+
;*   | Seg #1, Monitor                  |
;*   | $4000   (Possiby ROM)            |
;*   +----------------------------------+
;*   | Seg #2, Read/Write Window        |
;*   | $8000                            |
;*   +----------------------------------+
;*   | Seg #3, Normal                   |
;*   | $c000                            |
;*   +----------------------------------+
;*
;* STORAGE
;*   +----------------------------------+
;*   | $0000-$0400 zp + stack+ state    |
;*   | $0400-$0800 Screen RAM           |
;*   +----------------------------------+
;*
;*   $22     temporary (directory)
;*   $39     temporary
;*   $3a     temporary		
;*   $a3/$a4 pointer for banking load store/high-addrs for 'TL'
;*   $a5     temporary Y for banking load store/high-len for 'TL'
;*   $c1/$c2 data pointer
;*   $c3/$c4 data pointer
;*   $c9/$ca end pointer for 'T' and 'TL'
;*
;******

;* configuration
HAVE_DTV_V2	equ	1


DEFAULT_MEM	equ	$07

	seg.u	stack_mon
	org	$0120
;**************************************************************************
;*
;* SECTION  stack_mon
;*
;* DESCRIPTION
;*   Stack area allocation
;*
;******
;* storage
StartStore:
;***
PCStore:
	ds.w	1
SRStore:
	ds.b	1
ACStore:	
	ds.b	1
XRStore:	
	ds.b	1
YRStore:
	ds.b	1
SPStore:
	ds.b	1
BRK1Store:
	ds.w	1
MemoryStore:
	ds.b	1
BankStore:
	ds.b	1
RomStore:
	ds.b	1
ModeStore:
	ds.b	1
RetStore:
	ds.w	1
;***
EndStore:	
;***
	echo	"[monitor] stack store:",StartStore,"-",EndStore
	
	seg	code
;**************************************************************************
;**
;** ResetEntry2 and ResetEntry3 (hooks)
;**
;******
ResetEntry2:
	jsr	clear_store
	stx	RetStore
	sty	RetStore+1
 	bra	NMIEntry	

ResetEntry3:
	jsr	clear_store
	stx	RetStore
	sty	RetStore+1
	lda	#$07		;Normal Mem
	sta	$01
	tsx
	stx	SPStore		;Save the stack pointer
	lda	#DEFAULT_MEM
	sta	MemoryStore		;Default memory
	lda	#DEFAULT_REG8
	sta	RomStore

	jsr	$fda3
	ldx	#$c0
	stx	$9d
	jsr	init_system2
	bra	entry2

;**************************************************************************
;**
;** Reset and NMI entry points
;** Initialize monitor
;**
;******
ResetEntry:
	ldx	#8
	stx	$d016
	jsr	$fda3		;IOINIT
	ldx	#$fa
	txs		;FixaStack
	ldx	#$00
	stx	$d011	;Blankaskrm
	lda	#$06
	sta	$d020	;Bl.

	jsr	clear_store

NMIEntry:
	lda	#$07		;Normal Mem
	sta	$01
	lda	#$a0
	sta	$0284		;Memsiz = $a000 (End of mem)

	lda	#$04
	sta	$0288		;Screen at $0400

	tsx
	stx	SPStore		;Save the stack pointer
	lda	#DEFAULT_MEM
	sta	MemoryStore		;Default memory
	lda	#DEFAULT_REG8
	sta	RomStore
	
RestoreEntry:
	jsr	init_system

	jsr	init_screen
	jsr	print_starttext

entry2:
	jsr	re_InstallBRKNMI

	lda	#"S"
	jmp	DumpRegsGoMain

clear_store:
	stx	$c3
	ldx	#0
	txa
cls_lp1:
	sta	StartStore,x		;Empty stack area
	inx
	cpx	#(EndStore-StartStore)
	bne	cls_lp1
	ldx	$c3
	rts

init_system:
	jsr	clone_fd50_lite

init_system2:
	lda	#255
	sta	$028a		;enable KeyRepeat

	jsr	InstallLoadSavePatch

	jmp	clone2_fd15
;	rts

init_screen:
	jsr	$ff5b
	lda	#0
	jmp	SetColors
; 	rts

;**************************************************************************
print_starttext:
	ldx	#0
pst_lp1:
	lda	Startup_MSG,x
	jsr	$ffd2
	inx
	cpx	#[Startup_MSGEND-Startup_MSG]
	bne	pst_lp1
	rts
	
Startup_MSG:
	dc.b	147
	dc.b	"        DTVMON 1.2 (2008-02-17)",13
	dc.b	"  DONE IN 2005-2008 BY DANIEL KAHLIN.",13
	dc.b	"   BASED ON CCSMON BY PER H SUNDELL,",13
	dc.b	"     AND HESMON 2.0 BY T PETERSON.",13
Startup_MSGEND:

;**************************************************************************
re_InstallBRKNMI:
	ldx	#0
ribn_lp1:
	lda	ribn_Handler_st,x
	sta	ribn_Handler,x
	inx
	cpx	#ribn_Handler_end-ribn_Handler_st
	bne	ribn_lp1

	sir	$d2
	sty	ribn_sm1+1
	sir	$12

	lda	#>ribn_Handler
	sta	$0317
	sta	$0319
	lda	#<ribn_BRKEntry
	sta	$0316
	lda	#<ribn_NMIEntry
	sta	$0318

; preserve first byte of basic and initialize
	lda	$0800
	pha
	jsr	$e3bf	;initBasic
	pla
	sta	$0800

; set CHRGET/CHRGOT to $0200 (direct mode)
	lda	#$02
	sta	$7b
	rts

	
	echo	"[monitor] BRKNMI:",BRKNMIEntry,"-",BRKNMIEntry2
	if	[>BRKNMIEntry]!=[>BRKNMIEntry2]
	echo	"[monitor] BRKNMI same page failed!"
	err
	endif
ribn_Handler_st:
	rorg	EndStore
ribn_Handler:
ribn_NMIEntry:
	pha
	lda	#<BRKNMIEntry2
	sta	rh_sm1+1
	pla
ribn_BRKEntry:
	sir	$d8
ribn_sm1:
	ldy	#$06
	ldx	#DEFAULT_REG8
	sir	$12
rh_sm1:
	jmp	BRKNMIEntry
EndBreakHandler:
	rend
ribn_Handler_end:
	echo	"[monitor] break handler:",ribn_Handler,"-",EndBreakHandler
	
;**************************************************************************





;**************************************************************************
Commands:
	dc.b	"A"&$3f		;Assemble
	dc.b	"B"&$3f		;Set breakpoint!
	dc.b	"C"&$3f		;Compare
	dc.b	"D"&$3f		;Disassemble
	dc.b	"Q"&$3f		;Quicktrace
	dc.b	"F"&$3f		;Fill
	dc.b	"G"&$3f		;Go
	dc.b	"H"&$3f		;Hunt

	dc.b	"I"&$3f		;Ascii dump
	dc.b	"L"&$3f		;Load
	dc.b	"M"&$3f		;Hex dump
	dc.b	"N"&$3f		;New locate
	dc.b	"O"&$3f		;Disk command
	dc.b	"R"&$3f		;DumpRegs
	dc.b	"S"&$3f		;Save
	dc.b	"T"&$3f		;Transfer

	dc.b	"U"&$3f		;ColorTest
	dc.b	"W"&$3f		;Walk
	dc.b	"X"&$3f		;Exit
	dc.b	"#"&$3f		;Decimal to Hex
	dc.b	"$"&$3f		;Hex to Decimal
	dc.b	"*"&$3f		;Memory control
	dc.b	";"&$3f		;Change regs
	dc.b	":"&$3f		;Enter hexdata

	dc.b	"%"&$3f		;Binary to Hex and Decimal
	dc.b	"'"&$3f		;Enter Asciidata
	dc.b	"-"&$3f		;Hex subtract
	dc.b	"+"&$3f		;Hex add
	dc.b	","&$3f		;Enter CodeData
	dc.b	"E"&$3f		;Install breakhandler.
	dc.b	"@"&$3f		;Get diskstatus
	dc.b	"P"&$3f		;Set default output device

Commands_End:
;***
CommandPointers:
	dc.w	Command_A
	dc.w	Command_B
	dc.w	Command_C
	dc.w	Command_D
	dc.w	Command_Q
	dc.w	Command_F
	dc.w	Command_G
	dc.w	Command_H
	dc.w	Command_I
	dc.w	Command_L
	dc.w	Command_M
	dc.w	Command_N
	dc.w	Command_O
	dc.w	Command_R
	dc.w	Command_S
	dc.w	Command_T
	dc.w	Command_U
	dc.w	Command_W
	dc.w	Command_X
	dc.w	Command_Staket
	dc.w	Command_Dollar
	dc.w	Command_Star
	dc.w	Command_Semicolon
	dc.w	Command_Colon
	dc.w	Command_Percent
	dc.w	Command_Fnutt
	dc.w	Command_Minus
	dc.w	Command_Plus
	dc.w	Command_Comma
	dc.w	Command_E
	dc.w	Command_AlfaSlang
	dc.w	Command_P
;**************************************************************************


;**************************************************************************
;**
;** Dah real mainloop!
;**
;******
MainLoopBig:
	lda	#$01
	sta	$cc	;Vanlig cursor off!
	lda	$032b
	cmp	#$f1
	beq	sii_skp1	;Standard getin=skip

	jsr	$ffe4	;Get char.
sii_skp1:
	lda	#$00
	sta	$c6	;Tm charbuffer
	jsr	$ffe7	;Close all

MainLoopSmall:
	lda	$d1
	sta	$f3
	ldy	$d3
	lda	($d1),y		;Invertera kursor
	eor	#$80
	sta	($d1),y
	lda	$d2
	and	#$03
	ora	#$d8
	sta	$f4
	lda	($f3),y
	sta	$0287		;Spar teckenfrg
	lda	$0286
	sta	($f3),y		;Lagra cursorfrg
	ldx	#$fa
	txs		;Stackreset
ml_lp1:
	jsr	$ffe4	;Hmta tecken
	beq	ml_lp1
	sta	$3a	;Spara nuvarande tecken

	ldy	$d3
	lda	$0287
	sta	($f3),y	;terstll tecken under cursorn
	lda	($d1),y
	eor	#$80
	sta	($d1),y

	lda	$3a
	sty	$3a
	cmp	#13	;CR?
	beq	ml_skp1	;Ja, Kolla rad!
	jsr	CheckFKeys
	bcc	MainLoopSmall	;Nsta tecken!

ml_skp1:
	ldy	#$ff
	ldx	#$00
	jsr	GetAsciiChar
	cpy	#40
	bcs	CRMainLoopBig	; nothing found, skip check
ml_lp2:
	sta	$0200
	cmp	Commands,x
	beq	ml_Execute
	inx
	cpx	#Commands_End-Commands
	bne	ml_lp2
;Syntaxfel!
ErrMainLoopBig:
	lda	#"?"
	jsr	$e716
CRMainLoopBig:
	lda	#13	
	jsr	$e716
	jmp	MainLoopBig	;Looopa!


;*** Utfr Kommando! ***
ml_Execute:
	txa
	asl
	tax
	lda	CommandPointers,x
	sta	$c3
	lda	CommandPointers+1,x
	sta	$c4
	jmp	($00c3)		;Jeronimooooooo!
;**************************************************************************

;*** Fetch Screenbyte ***
GetScreenChar:
	tya
	tax
	clc
	adc	#$06
	tay
	lda	($d1),y
	pha
	txa
	tay
	pla
	rts
;**************************************************************************


;**************************************************************************
;**
;** Command 'X' - Exit monitor.
;**
;******
Command_X:
	jsr	GetAsciiChar
	cmp	#"C"&$3f
	bne	cmx_skp1
	jmp	clone_e394	;XC: Resetta basic!

cmx_skp1:
	cmp	#"Z"&$3f
	beq	cmx_skp2

	lda	RetStore
	ora	RetStore+1
	bne	cmx_skp3
	jmp	clone_e37b	;X: Hoppa ut till basic.
cmx_skp3:
	jmp	go_ind_retstore

cmx_skp2:
	lda	#$0e
	sta	$d020
	sta	$0286
	lda	#$04
	sta	$0288
	lda	#$06
	sta	$d021
	sei
	jsr	$ff5b
	jsr	clone2_fd15
	cli
	jmp	clone_e394	;XZ: Resetta Basic och video!
;**************************************************************************

;**************************************************************************
;**
;** Command 'R' - Dump registers
;**
;******
Command_R:
	ldx	#0
cmr_lp1:
	lda	RegLabels_MSG,x
	jsr	$ffd2
	inx
	cpx	#RegLabels_end-RegLabels_MSG
	bne	cmr_lp1
	jsr	cmr_DumpRegs
	jmp	MainLoopBig

cmr_DumpRegs:
	lda	PCStore+1		
	jsr	PrintHexByte
	lda	PCStore
	jsr	PrintHexByte		;Dump ProgramCounter

	lda	$0315
	jsr	PrintSpaceHexByte	;Dump IRQ-ptr
	lda	$0314
	jsr	PrintHexByte

	lda	SRStore
	jsr	PrintSpaceHexByte	;Dump SR

	lda	ACStore
	jsr	PrintSpaceHexByte	;Dump AC

	lda	XRStore
	jsr	PrintSpaceHexByte	;Dump XR

	lda	YRStore
	jsr	PrintSpaceHexByte	;Dump YR

	lda	SPStore			;Dump SP
	jsr	PrintSpaceHexByte

	lda	BRK1Store+1
	jsr	PrintSpaceHexByte	;Dump BRK1
	lda	BRK1Store
	jsr	PrintHexByte

	lda	MemoryStore
	jsr	PrintSpaceHexByte	;Dump Memory

	lda	#" "
	jsr	$ffd2
	lda	RomStore
	and	#%00110000
	php
	lda	BankStore
	lsr
	lsr
	plp
	bne	cdr_skp1
	ora	#$80			; Rom
cdr_skp1:
	jsr	PrintSpaceHexByte	;Dump Bank


	jsr	clone_ab3f
	jsr	read_d03f
	jsr	PrintHexNybble
	
	lda	#13
	jmp	$ffd2
;**************************************************************************


;*** PrintSpaceHexByte ***
PrintSpaceHexByte:
	pha
	lda	#" "
	jsr	$ffd2
	pla
;*** PrintHexByte ***
PrintHexByte:
	pha
	lsr
	lsr
	lsr
	lsr
	jsr	PrintHexNybble
	pla
; fall through
;*** Skriv ut nybble i acc ***
PrintHexNybble:
	and	#$0f
	clc
	adc	#$30
	cmp	#$3a
	bcc	phn_skp1
	clc
	adc	#$07
phn_skp1:
	jmp	$ffd2

;*** Hmta byte till Acc ***
GetHexByte:
	jsr	GetHexNybble
	asl
	asl
	asl
	asl
	sta	$3a
	jsr	GetHexNybble
	ora	$3a
	sta	$3a
	rts

;*** Hmta word till $c3,$c4... Acc=$c3
GetHexWord:
	jsr	GetHexByte
	sta	$c4
	jsr	GetHexByte
	sta	$c3
	rts

;*** Hmta nybble till Acc ***
GetHexNybble:
	jsr	GetAsciiChar
	cmp	#$30
	bcs	ghn_skp1
	clc
	adc	#$39
ghn_skp1:
	sec
	sbc	#$30
	rts

;**************************************************************************
RegLabels_MSG:
	dc.b	13,"  PC  IRQ  SR AC XR YR SP BRK1 MEM BN X"
	dc.b	13,";"
RegLabels_end:

Memory_MSG:
	dc.b	" MEMORY"
Memory_end:
	
;**************************************************************************
;**
;** Command '$' - convert Hex to decimal
;**
;******
Command_Dollar:	
	jsr	GetHexWord
	lda	#" "
	jsr	$ffd2
	lda	$c4
	ldx	$c3
	jsr	clone_bdcd	;Skriv decimalt
CRMainLoopSmall:
	lda	#13
	jsr	$ffd2
	jmp	MainLoopSmall


;**************************************************************************


;**************************************************************************
;**
;** Command '#' - convert decimal to hex
;**
;******
Command_Staket:			;Decimal to Hex!
	lda	#$00
	sta	$c1
	sta	$c2
cmst_lp1:
	jsr	GetAsciiChar
	sec
	sbc	#$30
	cmp	#$0a
	bcs	cmst_skp2
	pha
	asl	$c1
	rol	$c2
	lda	$c2
	pha
	lda	$c1
	pha
	asl
	rol	$c2
	asl
	rol	$c2
	sta	$c1
	pla
	clc
	adc	$c1
	sta	$c1
	pla
	adc	$c2
	sta	$c2
	pla
	clc
	adc	$c1
	sta	$c1
	bcc	cmst_skp1
	inc	$c2
cmst_skp1:
	jmp	cmst_lp1
cmst_skp2:
	jsr	clone_ab3f
	lda	#"$"
	jsr	$ffd2
	lda	$c2
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte
	jmp	CRMainLoopSmall
;**************************************************************************


;**************************************************************************
NoFKey:
	jsr	AutoScroller
	jmp	$e716
;**************************************************************************

;**************************************************************************
GetWordOrPC:
	jsr	GetHexWord
	cpy	#$27
	bcc	gwop_ex1
	lda	PCStore
	sta	$c3
	lda	PCStore+1
	sta	$c4
gwop_ex1:
	rts
;**************************************************************************
SetTimer:
	sta	$dd04		;Stt nmi-timer
	lda	#$08
st_lp1:
	bit	$d012	;Se upp fr char fetch!
	beq	st_lp1
st_lp2:
	bit	$d012
	bne	st_lp2
	rts
;**************************************************************************


;**************************************************************************
;**
;** Command '-' - Do hex subtract.
;**
;******
Command_Minus:			;hex subtract!
	jsr	GetTwoHexWords
	sec
	lda	$c1
	sbc	$c3
	sta	$c1
	lda	$c2
	sbc	$c4
	sta	$c2
	jmp	PrintDecHex
;**************************************************************************


;**************************************************************************
;**
;** Command '+' - Do hex add.
;**
;******
Command_Plus:			;hex add!
	jsr	GetTwoHexWords
	clc
	lda	$c1
	adc	$c3
	sta	$c1
	lda	$c2
	adc	$c4
	sta	$c2
	jmp	PrintDecHex
;**************************************************************************


;*** GetTwoHexWords $c1/$c2 och $c3/$c4 acc=$c3 ***
GetTwoHexWords:
	jsr	GetHexWord
	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2
	jsr	GetHexWord
	rts

;*** Hmta hsta tecken skippa space! Acc=Tecken ***
GetAsciiChar:
	iny
	cpy	#$28
	bcs	gac_skp1
	lda	($d1),y
	cmp	#" "
	beq	GetAsciiChar
	rts
gac_skp1:
	lda	$0200
	cmp	#$23
	beq	gac_skp2
	lda	#"0"
	rts
gac_skp2:
	lda	#" "
	rts

;**************************************************************************


;**************************************************************************
;**
;** Command 'F' - Fill memory area.
;**
;******
Command_F:			;Command fill!
	jsr	GetTwoHexWords
	jsr	GetHexByteCheckEnd
	jsr	fill_mem_c1_c3
	jmp	CRMainLoopSmall

fill_mem_c1_c3:
	sta	$3a
	ldy	#0
cfl_lp1:
	lda	$3a
	jsr	PutByte_c1
	sei
	inc	$c1
	bne	cfl_skp1
	inc	$c2
cfl_skp1:
	ldx	$c1
	cpx	$c3
	bne	cfl_lp1
	ldx	$c2
	cpx	$c4
	bne	cfl_lp1

	lda	$3a
	jsr	PutByte_c1
	rts
	


;**************************************************************************


QuoteOffMainLoopSmall:
	lda	#$00
	sta	$d4	;Quotemode off!
	jmp	MainLoopSmall


;**************************************************************************



;**************************************************************************
;**
;** Command '%' - convert binary to Hex and decimal
;**
;******
Command_Percent:		;Binary to hex+decimal
	lda	#$00
	sta	$c1
	ldx	#$00
	stx	$c2
cmpe_lp1:
	jsr	GetAsciiChar
	cpy	#$27
	bcs	PrintDecHex
	lsr
	rol	$c1
	rol	$c2
	inx
	bne	cmpe_lp1

;*** Dump $c1,$c2 in hex and decimal ***
PrintDecHex:
	jsr	clone_ab3f		;space

	lda	#"$"
	jsr	$ffd2
	lda	$c2
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte

	jsr	clone_ab3f		;space

	lda	$c2
	ldx	$c1
	jsr	clone_bdcd
	jmp	CRMainLoopSmall
;**************************************************************************


GetHexByteCheckEnd:
	jsr	GetHexByte
CheckEnd:
	cpy	#$28
	bcs	ghbce_ex1
	rts
ghbce_ex1:
	jmp	CRMainLoopBig

;**************************************************************************



;**************************************************************************
;**
;** Command 'G' - Jump to code.
;**
;******
Command_G:			;GO!
	jsr	GetWordOrPC

	ldx	#cmg_GoLen-1
cmg_lp1:
	lda	cmg_GoSt,x
	sta	cmg_Go,x
	dex
	bpl	cmg_lp1

	lda	$c3
	sta	go_sm1+1
	lda	$c4
	sta	go_sm1+2

	sei
	ldx	SPStore
	txs
	lda	SRStore
	pha
	ldx	XRStore
	ldy	YRStore
	lda	MemoryStore
	sta	$01
	lda	ACStore
	sir	$d8
	jmp	cmg_Go

;*** Go routine to be placed at $0110 ***
cmg_GoSt	equ	.
	rorg	$0110
cmg_Go:
	ldy	#$01		; map $004000 -> $4000
	ldx	#%01010101	; in ram
	sir	$12
	plp			; pop status register from stack
go_sm1:
	jmp	go_sm1
	rend
cmg_GoLen	equ	.-cmg_GoSt
;**************************************************************************


;**************************************************************************
;**
;** Command ';' - Change registers...
;**
;******
Command_Semicolon:		;Change Regs...
	jsr	GetHexWord
	sta	PCStore
	lda	$c4		;Set PC
	sta	PCStore+1

	jsr	GetHexWord
	sei
	sta	$0314
	lda	$c4		;Set IRQ
	sta	$0315

	jsr	GetHexByte
	sta	SRStore		;Set SR

	jsr	GetHexByte
	sta	ACStore		;Set AC

	jsr	GetHexByte
	sta	XRStore		;Set XR

	jsr	GetHexByte
	sta	YRStore		;Set YR

	jsr	GetHexByte
	sta	SPStore		;Set SP

	jsr	GetHexWord
	sta	BRK1Store
	lda	$c4
	sta	BRK1Store+1	;Set BRK1

	jsr	GetHexByte
	sta	MemoryStore	;Set Memory

	jsr	GetHexByte
	pha
	asl
	asl
	sta	BankStore	;Set Bank
	pla
	ldx	#DEFAULT_REG8
	and	#$80
	beq	csc_skp1
	ldx	#DEFAULT_REG8 & %01000101
csc_skp1:
	stx	RomStore	;Set Ram or Rom

	jsr	GetHexNybble
	sta	$d03f

	cli
	ldy	#1
	sty	$d3		;Cursor pos 1
	jsr	cmr_DumpRegs

	jmp	MainLoopSmall
;**************************************************************************


;**************************************************************************
;**
;** Command '*' - Change memory mode.
;**
;******
Command_Star:				;Memorycontrol

	lda	MemoryStore	;Decrease Memory 
	and	#$07		;Low 3bits
	tax
	dex
	txa
	and	#$07
	sta	$c1
	lda	MemoryStore
	and	#$f8
	ora	$c1
	sta	MemoryStore

	ldy	#$01
	sty	$d3		;Cursor pos 1
	dey
cmsr_lp1:
	lda	Memory_MSG,y
	jsr	$ffd2
	iny
	cpy	#Memory_end-Memory_MSG
	bne	cmsr_lp1
	lda	MemoryStore
	jsr	PrintSpaceHexByte
	lda	#$00
	sta	$d3
	jmp	MainLoopSmall
;**************************************************************************


;**************************************************************************
;**
;** Command '@' - Get diskstatus.
;**
;******
Command_AlfaSlang:			;Get Diskstatus
	ldx	#$08
	lda	#$0f
	tay
	jsr	$ffba
	lda	#$00		;Open error channel
	jsr	$ffbd
	jsr	$ffc0
	ldx	#$0f
	jsr	$ffc6

	lda	#13
	jsr	$ffd2

cmas_lp1:
	jsr	$ffcf
	jsr	$ffd2		;Get Status string and print
	cmp	#13
	bne	cmas_lp1

	jsr	$ffe7		;CLALL

	jmp	MainLoopBig
;**************************************************************************


;**************************************************************************
;**
;** Command 'O' - Send diskcommand.
;**
;******
Command_O:			;SendDiskcom
	jsr	GetFileName		;Get and Set Name
	lda	#$0f
	tay
	ldx	#$08
	jsr	$ffba
	jsr	$ffc0		;Send command
	jsr	$ffe7
	jmp	CRMainLoopBig
;**************************************************************************


;**************************************************************************
;**
;** Command 'P' - Set default output device.
;**
;******
Command_P:
	jsr	GetHexByte
	tax
	ldy	#$ff
	tya
	jsr	$ffba
	lda	#$00
	jsr	$ffbd	;Set Standard out!
	jsr	$ffc0
	ldx	#$ff
	jsr	$ffc9
	lda	#13
	jsr	$e716
	jmp	MainLoopSmall
;**************************************************************************


;**************************************************************************
;**
;** Command 'M' - Dumpmemory.
;**
;******
Command_M:			;Show memdump!
	jsr	GetTwoHexWords
	cpy	#$26
	bcc	cmm_skp1
DumpMemoryLine:
	lda	$c1
	sta	$c3	;If no End, set End to Start
	lda	$c2
	sta	$c4
cmm_skp1:

cmm_lp2:
	lda	#13
	jsr	$ffd2
	lda	#":"
	jsr	$ffd2
	lda	$c2		;Print Line address
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte

	ldy	#0
cmm_lp3:
	jsr	GetByte_c1
	sta	$0220,y		;Get 8 byte chunk from mem!
	jsr	PrintSpaceHexByte	;And print with spaces
	iny
	cpy	#8
	bne	cmm_lp3

	jsr	clone_ab3f	;PrintSpace

	ldy	#$00
cmm_lp4:
	lda	#$01
	sta	$d4	;Quotemode
	lda	$0220,y
	jsr	PrintCharClean		;and print
	iny
	cpy	#$08
	bne	cmm_lp4

	lda	#$00
	sta	$d3

	ldx	#$08
	bne	cmm_skp3
;***
cmm_lp5:
	inc	$c1
	bne	cmm_skp2
	inc	$c2
cmm_skp2:
	dex
	beq	cmm_skp4
cmm_skp3:
	lda	$c4
	cmp	$c2	;Past End... stick!
	bcc	cmm_ex1
	bne	cmm_lp5
	lda	$c1
	cmp	$c3
	bcc	cmm_lp5	;PastEnd... stick!
cmm_ex1:
	jmp	QuoteOffMainLoopSmall
;***
cmm_skp4:
	lda	$91
	bmi	cmm_lp2
	bpl	cmm_ex1





PrintCharClean:
	pha
	sta	$d7
	txa
	pha
	tya
	pha
	lda	#$00
	sta	$d0
	ldy	$d3
	lda	$d7
	bpl	pcc_skp2
	cmp	#$a0
	bcc	pcc_skp1
	jmp	pcc_skp6
pcc_skp1:
	ora	#$40
	jmp	$e697
pcc_skp2:
	cmp	#$20
	bcs	pcc_skp3
	jmp	$e697
pcc_skp3:
	cmp	#$60
	bcc	pcc_skp4
	and	#$df
	bne	pcc_skp5
pcc_skp4:
	and	#$3f
pcc_skp5:
	jmp	$e693
;**************************************************************************


;**************************************************************************
;**
;** Command 'I' - Dump ascii.
;**
;******
Command_I:			;Show Lines!!
	jsr	GetTwoHexWords
	cpy	#$26
	bcc	cmi_skp1
DumpAsciiLine:
	lda	$c1
	sta	$c3		;Om inget slut, d slut=Start
	lda	$c2
	sta	$c4
cmi_skp1:
	
cmi_lp2:
	lda	#13
	jsr	$ffd2
	lda	#"'"
	jsr	$ffd2
	lda	$c2		;Print line address
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte
	lda	#" "
	jsr	$ffd2

	ldy	#$00
cmi_lp3:
	lda	#$01
	sta	$d4	;Quote mode on
	jsr	GetByte_c1
	jsr	PrintCharClean		;Skriv $20 bytes
	iny
	cpy	#$20
	bne	cmi_lp3

	lda	#$00
	sta	$d3
	ldx	#$20
	bne	cmi_skp3
;***
cmi_lp4:
	inc	$c1
	bne	cmi_skp2
	inc	$c2
cmi_skp2:
	dex
	beq	cmi_skp4
cmi_skp3:
	lda	$c4
	cmp	$c2
	bcc	cmi_ex1
	bne	cmi_lp4
	lda	$c1
	cmp	$c3
	bcc	cmi_lp4
cmi_ex1:
	jmp	QuoteOffMainLoopSmall

cmi_skp4:
	lda	$91
	bmi	cmi_lp2
	bpl	cmi_ex1
;**************************************************************************


;**************************************************************************
;**
;** Command ':' - Enter hexdata.
;**
;******
Command_Colon:		;Enter hexdata!
	jsr	GetHexWord

	ldx	#0
cmco_lp2:
	jsr	GetHexByte
	cpy	#$26
	bcc	cmco_skp1
	bcs	cmco_ex1
cmco_skp1:
	sty	$c1
	stx	$c2
	ldy	$c2
	jsr	cmco_Update
	ldy	$c1
	inx
	cpx	#$08
	bne	cmco_lp2

	lda	$c3
	clc
	adc	#$08
	sta	$c3
	bcc	cmco_skp2
	inc	$c4
cmco_skp2:
	lda	#13
	jsr	$ffd2
	lda	#":"
	jsr	$ffd2
DumpAddressMain:
	lda	$c4
	jsr	PrintHexByte
	lda	$c3
	jsr	PrintHexByte
	inc	$d3		;Cursor pos
	jmp	MainLoopSmall
cmco_ex1:
	jmp	CRMainLoopSmall


cmco_Update:
	jsr	cmco_UpdateChar
	jsr	PutByte_c3
	rts

cmco_UpdateChar:
	pha
	tya
	clc
	adc	#$1e
	sta	$d3
	pla
	jmp	PrintCharClean
;**************************************************************************


;**************************************************************************
pcc_skp6:
	and	#$7f
	jmp	$e691	;putchar...
;**************************************************************************
GetCharFromScreen:
	sta	$d7
	tya
	pha
	txa
	pha
	dec	$d3
	ldy	#$00
	sty	$c8
	iny
	sty	$d4
	lda	$d7
	jmp	$e640
;**************************************************************************


;**************************************************************************
;**
;** Command "'" - Enter ascii data.
;**
;******
Command_Fnutt:			;Enter AsciiData
	ldy	#0
	jsr	GetHexWord

	lda	$d3
	cmp	#$06		;Nothing typed... exit!
	beq	cmco_ex1

	ldy	#0
cmfn_lp2:
	jsr	GetScreenChar
	jsr	ConvertAscii
	jsr	PutByte_c3
	iny
	cpy	#$20
	bne	cmfn_lp2
	lda	$c3
	clc
	adc	#$20
	sta	$c3
	bcc	cmfn_skp1
	inc	$c4
cmfn_skp1:
	lda	#13
	jsr	$ffd2
	lda	#"'"
	jsr	$ffd2
	jmp	DumpAddressMain
;**************************************************************************


;**************************************************************************
;**
;** Command 'U' - ColorTest!
;**
;******
Command_U:			;ColorTest!
	lda	$0286
	ldx	#$00
	ldy	#$00
cmu_lp1:
	sta	$d800,y
	sta	$d900,y
	sta	$da00,y
	sta	$db00,y
	iny
	bne	cmu_lp1
	sec
	sbc	#$01
	and	#$0f
	inx
	cpx	#$11
	bne	cmu_lp1

	jsr	clone_ab3f		;Space
	lda	#"O"
	jsr	$ffd2
	lda	#"K"
	jsr	$ffd2
	lda	#"."
	jsr	$ffd2
	jmp	CRMainLoopSmall
;**************************************************************************


;**************************************************************************
;**
;** Command 'TL' - Transfer memory long.
;**
;******
Command_TL:
	jsr	GetHexByte
	sta	$a3
	jsr	GetHexWord
	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2
	jsr	GetHexByte
	sta	$a5
	jsr	GetHexWord
	lda	$c3
	sta	$c9
	lda	$c4
	sta	$ca
	jsr	GetHexByte
	sta	$a4
	jsr	GetHexWord
	jsr	CheckEnd

; source = $c1/$c2/$a3-$c9/$ca/$a5  target=$c3/$c4/$a4

	lda	$c9
	sec
	sbc	$c1
	sta	$c9
	lda	$ca
	sbc	$c2
	sta	$ca
	lda	$a5
	sbc	$a3
	and	#%00111111	; don't care about ram/rom differences
	sta	$a5
; $c9/$ca/$a5=len-1

	lda	#%00001101	; Forward DMA.
	sta	$3a

; is target less than source?
	lda	$c1
	cmp	$c3
	lda	$c2
	sbc	$c4
	lda	$a3
	sbc	$a4
	bcs	ctl_forward	; yes, go forwards

; reverse the range
	lda	$c1
	clc
	adc	$c9
	sta	$c1
	lda	$c2
	adc	$ca
	sta	$c2
	lda	$a3
	adc	$a5
	sta	$a3
	lda	$c3
	clc
	adc	$c9
	sta	$c3
	lda	$c4
	adc	$ca
	sta	$c4
	lda	$a4
	adc	$a5
	sta	$a4

	lda	#%00000001	; Backward DMA.
	sta	$3a
			
ctl_forward:
; fix up len.
	inc	$c9
	bne	ctl_skp1
	inc	$ca
	bne	ctl_skp1
	inc	$a5
ctl_skp1:
; $c9/$ca/$a5=len
	
	ldy	#1
	sty	$d03f
	jsr	clear_dma
	sty	$d306		; source step
	sty	$d308		; dest step
	
ctl_lp1:
	lda	$c1
	sta	$d300
	lda	$c2
	sta	$d301
	lda	$a3
	jsr	adjust_dmaaddr
	sta	$d302
	lda	$c3
	sta	$d303
	lda	$c4
	sta	$d304
	lda	$a4
	and	#%00111111
	ora	#$40		;Target is always RAM.
	sta	$d305

	lda	$a5
	beq	ctl_skp3	;<$010000 left, only one operation needed
; more than $ffff, move $010000 at once.
	lda	#0
	sta	$d30a
	sta	$d30b
	jsr	do_dma_clean
	dec	$a5		; len-=$010000
	lda	$3a
	cmp	#%00001101	; forward?
	beq	ctl_skp2	; yes, increment addr.
	dec	$a3		; source-=$010000
	dec	$a4		; dest-=$010000
	bra	ctl_lp1
ctl_skp2:
	inc	$a3		; source+=$010000
	inc	$a4		; dest+=$010000
	bra	ctl_lp1
ctl_skp3:
	lda	$c9
	sta	$d30a
	ora	$ca
	beq	ctl_skp4	; both $c9/$ca=0, all copied, exit.
	lda	$ca
	sta	$d30b
	jsr	do_dma_clean

ctl_skp4:
	jsr	clear_dma
	ldy	#0
	sty	$d03f
	jmp	CRMainLoopSmall
	
; convert a high addr/bank into a DMA controller high addr.
adjust_dmaaddr:
	and	#%10111111
	bmi	aad_skp1
	ora	#%01000000
aad_skp1:
	and	#%01111111
	rts

do_dma_clean:
	lda	$3a
	sta	$d31f
ddc_lp1:
	lda	$d31f
	lsr
	bcs	ddc_lp1
	rts
	
;**************************************************************************
;**
;** Command 'T' - TransferMemory.
;** uses $c1,$c2,$c3,$c4,$c9,$ca,$d7
;**
;******
Command_T:		;Transfer!
	jsr	GetAsciiChar
	cmp	#"L" & $3f
	bne	cmt_skpchk
	jmp	Command_TL
cmt_skpchk:
	dey
	jsr	GetTwoHexWords
	sta	$c9
	lda	$c4	;End to $c9,$ca
	sta	$ca
	jsr	GetHexWord
	jsr	CheckEnd

	lda	#$00
	sta	$d7
	lda	$c4
	cmp	$c2
	bcc	cmt_skp2
	bne	cmt_skp1
	lda	$c3
	cmp	$c1
	bcc	cmt_skp2

;*** set backwardsmode ***
cmt_skp1:
	lda	$c9
	sec
	sbc	$c1
	tax
	lda	$ca
	sbc	$c2
	tay
	txa
	clc
	adc	$c3
	sta	$c3
	tya
	adc	$c4
	sta	$c4
	lda	$ca
	ldx	$c2
	stx	$ca
	sta	$c2
	lda	$c9
	ldx	$c1
	stx	$c9
	sta	$c1
	lda	#$01
	sta	$d7
cmt_skp2:
cmt_lp3:
	ldy	#0
	jsr	GetByte_c1		;CopyLoop!
	jsr	PutByte_c3
	lda	$c2
	cmp	$ca
	bne	cmt_skp3
	lda	$c1
	cmp	$c9
	bne	cmt_skp3
	jmp	CRMainLoopSmall
cmt_skp3:
	lda	$d7	;backwards?
	bne	cmt_skp6	;yes skip to decrease

;*** Increase $c1/$c2, $c3/$c4 ***
	inc	$c1
	bne	cmt_skp4
	inc	$c2
cmt_skp4:
	inc	$c3
	bne	cmt_skp5
	inc	$c4
cmt_skp5:
	jmp	cmt_lp3

;*** Decrease $c1/$c2, $c3/$c4 ***
cmt_skp6:
	lda	$c1
	bne	cmt_skp7
	dec	$c2
cmt_skp7:
	dec	$c1
	lda	$c3
	bne	cmt_skp8
	dec	$c4
cmt_skp8:
	dec	$c3
	jmp	cmt_lp3
;**************************************************************************
	

;**************************************************************************
;**
;** Command 'C' - Compare memory
;**
;******
Command_C:			;Compare!!
	jsr	GetTwoHexWords
	sta	$bd
	lda	$c4
	sta	$be
	jsr	GetHexWord

	lda	#13
	jsr	$ffd2
cmc_lp2:
	ldy	#0
	jsr	GetByte_c1
	sta	$d7
	jsr	GetByte_c3
	cmp	$d7
	beq	cmc_skp1
	lda	$c2
	jsr	PrintSpaceHexByte
	lda	$c1
	jsr	PrintHexByte
cmc_skp1:
	lda	$c2
	cmp	$be
	bne	cmc_skp2
	lda	$c1
	cmp	$bd
	bne	cmc_skp2
cmc_ex1:
	jmp	CRMainLoopSmall


cmc_skp2:
	inc	$c1
	bne	cmc_skp3
	inc	$c2
cmc_skp3:
	inc	$c3
	bne	cmc_skp4
	inc	$c4
cmc_skp4:
	lda	$91
	bmi	cmc_lp2
	bpl	cmc_ex1
;**************************************************************************



;**************************************************************************
;**
;** Kolla fknstangenter
;**
;** F1=Directory
;** F3=ChangeColors
;** F5=Type "L  08"
;** F8=ClearMem
;**
;******
CheckFKeys:
	cmp	#$85		;F1
	beq	cfk_F1
	cmp	#$86		;F3
	beq	cfk_F3
	cmp	#$87		;F5
	beq	cfk_F5
	cmp	#$8c		;F8
	beq	cfk_F8
	jmp	NoFKey
cfk_F1:
	jmp	ShowDirectory
cfk_F3:
	jmp	SetAltColors
cfk_F5:
	jmp	TypeLoad
cfk_F8:
	jmp	ClearMem
;**************************************************************************
ShowDirectory:
	jsr	$fda3
	lda	#147
	jsr	$ffd2
	lda	#$08
	tax
	ldy	#$00
	jsr	$ffba
	lda	#$01
	ldx	#<sdd_dirname
	ldy	#>sdd_dirname
	jsr	$ffbd
	jsr	$ffc0
	ldy	$90		;Status fel... ut!
	bne	sdd_fl1

	lda	#$08
	jsr	$ffb4
	lda	#$00
	jsr	$ff96

;***
	lda	#$05
sdd_lp1:
	sta	$22
sdd_lp2:
	jsr	$ffa5
	dec	$22
	bne	sdd_lp2
	sta	$22

	jsr	$ffa5
	ldx	$22
	ldy	$90		;Status fel... ut!
	bne	sdd_ex1

	jsr	clone_bdcd	;Print line
	jsr	clone_ab3f	;space
sdd_lp3:
	jsr	$ffa5
	jsr	$ffd2
	bne	sdd_lp3

	lda	#13
	jsr	$ffd2

	lda	$91
	bpl	sdd_ex1		;Break!

	lda	#$03
	bne	sdd_lp1
;***
sdd_fl1:
	lda	#13
	jsr	$ffd2
sdd_ex1:
	lda	#$08
	jsr	$ffab
	lda	#$08
	jsr	$ffc3
	lda	#$00
	sta	$c6
	rts
sdd_dirname:
	dc.b	"$"
;**************************************************************************
ColorTable	equ	cfg_dtvmon_colors1
	if	[cfg_dtvmon_colors2-cfg_dtvmon_colors1] != 3
	echo	"[monitor] color config not back to back!"
	err
	endif

SetAltColors:
	lda	#1
;***
SetColors:
	sta	$0200
	asl
	clc
	adc	$0200
	tax
	lda	ColorTable,x
	sta	$d020
	lda	ColorTable+1,x
	sta	$d021
	lda	ColorTable+2,x
	sta	646

	ldx	#0
seco_lp1:
	sta	$d800,x
	sta	$d900,x
	sta	$da00,x
	sta	$dae8,x
	inx
	bne	seco_lp1
	rts

;**************************************************************************
TypeLoad:
	ldy	#$00
	lda	#$0c
	sta	($d1),y
	iny
	lda	#$20
tl_lp1:
	sta	($d1),y
	iny
	cpy	#$05
	bne	tl_lp1

	ldy	#$18
	lda	#$30
	sta	($d1),y
	iny
	lda	#$38
	sta	($d1),y
	lda	#$20
tl_lp2:
	iny
	sta	($d1),y
	cpy	#$27
	bne	tl_lp2

	ldy	#$00
	lda	$0286
tl_lp3:
	sta	($f3),y
	iny
	cpy	#$28
	bne	tl_lp3
	ldy	#$00
	sty	$d3
	rts
;**************************************************************************
ClearMem:
	lda	#$00
	sta	$c1
	lda	#$08
	sta	$c2
	lda	#$ff
	sta	$c3
	sta	$c4
	lda	BankStore
	pha
	lda	MemoryStore
	pha
	lda	#$00
	sta	BankStore
	lda	#$04
	sta	MemoryStore
	lda	#$00
	jsr	fill_mem_c1_c3
	pla
	sta	MemoryStore
	pla
	sta	BankStore
	cli
	rts
;**************************************************************************



;**************************************************************************
GetNameOrSector:
	jsr	GetAsciiChar
	cmp	#34	;'"'
	bne	gnos_skp1
	jmp	DoLoadSave
gnos_skp1:
	dey
	jsr	GetHexWord
	sta	$c1
	lda	$c4
	sta	$c2
	jsr	GetHexWord
	cpy	#40-1
	bcc	gnos_ex1
	lda	#$ff
	sta	$c4
	lda	#$00
	sta	$c3
gnos_ex1:
	rts
;**************************************************************************


;**************************************************************************
;**
;** Command 'L' - Load file or diskblock
;**
;******
Command_L:			;Load
	jsr	GetNameOrSector

	ldx	#DiskMsg_len
cml_lp2:
	lda	DiskMsgSt,x
	sta	DiskMsg,x
	dex
	bpl	cml_lp2
	lda	#$31		;set Readsector
	sta	$0212

	jsr	SetTrackSector
	jsr	OpenCommandFiles
	jsr	SendReadWrite
	jsr	SendBPReset
	jsr	GetDiskBuffer
	jsr	CloseCommandFiles
	jmp	CRMainLoopSmall
;**************************************************************************


;**************************************************************************
;**
;** Command 'S' - Save file or diskblock
;**
;******
Command_S:			;Save
	jsr	GetNameOrSector

	ldx	#DiskMsg_len
cms_lp2:
	lda	DiskMsgSt,x
	sta	DiskMsg,x
	dex
	bpl	cms_lp2

	jsr	SetTrackSector
	jsr	OpenCommandFiles
	jsr	SendBPReset
	jsr	SendDiskBuffer
	jsr	SendReadWrite
	jsr	CloseCommandFiles
	jmp	CRMainLoopSmall
;**************************************************************************


;**************************************************************************
SetTrackSector:
	ldy	#0
	lda	$c2
	jsr	sts_SetByte
	lda	$c1
	iny
	lda	$c1
	jsr	sts_SetByte
	rts
sts_SetByte:
	pha
	and	#$f0
	lsr
	lsr
	lsr
	lsr
	jsr	sts_SetNybble
	pla
	and	#$0f
	jsr	sts_SetNybble
	rts
sts_SetNybble:
	clc
	adc	#$30
	sta	$0219,y
	iny
	rts
;***
OpenCommandFiles:
	jsr	$ffe7	;CLALL

	lda	#$0f
	ldx	#$08
	ldy	#$0f
	jsr	$ffba
	lda	#$00
	jsr	$ffbd
	jsr	$ffc0

	lda	#$08
	tax
	tay
	jsr	$ffba
	lda	#$01
	ldx	#$10
	ldy	#$02
	jsr	$ffbd
	jsr	$ffc0

	rts
;***
SendReadWrite:
	ldx	#$0f
	jsr	$ffc9
	ldx	#$00
srw_lp1:
	lda	$0211,x
	jsr	$ffd2
	inx
	cmp	#13
	bne	srw_lp1
	jsr	$ffcc	;CLCHN
	rts
;***
SendBPReset:
	ldx	#$0f
	jsr	$ffc9
	ldx	#$00
sbpr_lp1:
	lda	$0220,x
	jsr	$ffd2
	inx
	cmp	#13
	bne	sbpr_lp1
	jsr	$ffcc	;CLCHN
	rts
;***
CloseCommandFiles:
	jsr	$ffcc	;CLCHN
	lda	#$08
	jsr	$ffc3	;Close
	lda	#$0f
	jsr	$ffc3	;CLose
	jsr	$ffe7	;CLALL
	rts
;***
GetDiskBuffer:
	ldx	#$08
	jsr	$ffc6
	ldy	#$00
gdb_lp1:
	jsr	$ffcf
	jsr	PutByte_c3
	iny
	bne	gdb_lp1
	rts
;***
SendDiskBuffer:
	ldx	#$08
	jsr	$ffc9
	ldy	#$00
sdb_lp1:
	jsr	GetByte_c3
	jsr	$ffd2
	iny
	bne	sdb_lp1
	rts


;**************************************************************************
GetCharCheckQuote:
	lda	#34	; '"'
	iny
	cpy	#40
	bcs	gccq_ex1
	lda	($d1),y
	jsr	XLateChar
gccq_ex1:
	cmp	#34	; '"' 
	rts
;***
GetFileName:
	ldx	#0
gfn_lp1:
	jsr	GetCharCheckQuote
	bne	gfn_lp1
gfn_lp2:
	jsr	GetCharCheckQuote
	beq	gfn_ex1
	sta	$0201,x
	inx
	bne	gfn_lp2
gfn_ex1:
	stx	$b7
	lda	#$01
	sta	$bb
	lda	#$02
	sta	$bc
	rts
;**************************************************************************


;**************************************************************************
DoLoadSave:
	ldy	#$00
	jsr	GetFileName
	jsr	GetHexByte
	sta	$ba
	lda	#$01
	sta	$b9

	lda	$0200
	cmp	#"S"&$3f
	beq	dls_Save
;***
dls_Load:
	jsr	GetHexWord
	cpy	#$28
	bcs	dlsl_skp1
	dec	$b9
dlsl_skp1:
	ldx	$c3
	ldy	$c4
	lda	#$00
	jsr	clone_ffd5
	lda	$90
	and	#$10
	beq	dls_ex1
dls_fl1:
	jmp	ErrMainLoopBig
dls_ex1:
	jmp	CRMainLoopBig
;***
dls_Save:
	jsr	GetTwoHexWords	; Start + end
	cpy	#$28
	bcs	dls_fl1
; set load and save addresses $c1/$c2/$ac/$ad=start, $ae/$af=end+1
	ldx	#3
dlss_lp1:
	lda	$c1,x
	sta	$ac,x
	dex
	bpl	dlss_lp1
; attempt to get load addr
	jsr	GetHexWord	; $c3/$c4 = load addr
	cpy	#$28
	bcs	dlss_skp1
	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2
dlss_skp1:
	jsr	clone2_f5ed
dlss_ex1:
	jmp	CRMainLoopBig
;**************************************************************************


;**************************************************************************
DiskMsgSt:
	rorg	$0210
DiskMsg:
	dc.b	"#"
	dc.b	"U2: 8 0 18,00 ",13
	dc.b	"B-P:8 0",13
	rend
DiskMsg_len	equ .-DiskMsgSt
;**************************************************************************


;**************************************************************************
ConvertAscii:
	cmp	#$5e
	bne	coas_skp1
	lda	#$de
	inc	$d3
	bne	coas_skp2
coas_skp1:
	jsr	GetCharFromScreen
coas_skp2:
	sta	$c1
	cmp	#$a0
	bcs	coas_skp3
	rts
coas_skp3:
	cmp	#$c0
	php
	jsr	GetByte_c3
	sta	$c2
	plp
	bcs	coas_skp4
	eor	#$40
	cmp	$c1
	beq	coas_ex1
	lda	$c1
	rts
coas_ex1:
	lda	$c2
	rts
coas_skp4:
	lda	$c1
	cmp	$c2
	beq	coas_ex2
	eor	#$a0
coas_ex2:
	rts
;**************************************************************************
AutoScroller:
	sta	$39
	and	#$7f
	cmp	#$11
	bne	ascr_skp2
	lda	#$18
	sta	$c1
	ldx	#$00
	bit	$39
	bmi	ascr_skp1
	tax
ascr_skp1:
	stx	$02a5
	cpx	$d6
	beq	ascr_skp3
ascr_skp2:
	lda	$39
	rts
ascr_skp3:
	bit	$39
	bpl	ascr_skp4
	lda	#$13
	jsr	$e716
ascr_skp4:
ascr_lp1:
	lda	#13
	jsr	$e716
	lda	#145	;Cursor up!
	jsr	$e716

	ldy	#$ff
	jsr	GetAsciiChar

	cmp	#","
	beq	ascr_skp5
	cmp	#":"
	beq	ascr_skp5

ascr_lp2:
	cmp	#"'"
	beq	ascr_skp5
	lda	$39
	eor	#$80
	jsr	$e716
	dec	$c1
	bne	ascr_lp1
	ldx	$02a5
	dex
	stx	$d6
	lda	#$0d
	rts
ascr_skp5:
	sta	$0200

	jsr	GetHexWord
	cpy	#$27
	bcs	ascr_lp2
	ldx	$02a5
	dex
	stx	$d6
	lda	#13
	jsr	$e716
	bit	$39
	bmi	ascr_skp8	;Down!!

;*** Scroll Screenup! ***
	lda	#145	;Cursor up!
	jsr	$e716

	lda	$0200
	cmp	#","
	beq	ascr_skp6
	cmp	#"'"
	beq	ascr_skp7
	jmp	ScrollMemoryUp
ascr_skp6:
	jmp	ScrollDisasmUp
ascr_skp7:
	jmp	ScrollAsciiUp

;*** Scroll screendown! ***
ascr_skp8:
	lda	#$80
	sta	$d8
	jsr	$e981
	lda	$d9
	ora	#$80
	sta	$d9
	lda	#$00
	sta	$d8

	lda	#19	;Home
	jsr	$e716

	lda	$0200
	cmp	#","
	beq	ascr_skp9
	cmp	#"'"
	beq	ascr_skp10

	jmp	ScrollMemoryDown
ascr_skp9:
	jmp	ScrollDisasmDown
ascr_skp10:
	jmp	ScrollAsciiDown
;**************************************************************************
ScrollMemoryUp:
	lda	#$08
	jsr	AddToMemPtr
	jmp	DumpMemoryLine
;**************************************************************************
ScrollMemoryDown:
	lda	#$f8
	jsr	SubFromMemPtr
	jmp	DumpMemoryLine
;**************************************************************************
ScrollAsciiUp:
	lda	#$20
	jsr	AddToMemPtr
	jmp	DumpAsciiLine
;**************************************************************************
ScrollAsciiDown:
	lda	#$e0
	jsr	SubFromMemPtr
	jmp	DumpAsciiLine
;**************************************************************************
ScrollDisasmUp:
	lda	$c4
	sta	$c2
	lda	$c3
	sta	$c1
	lda	#17  ;Cursor Down
	jsr	$ffd2
	lda	#","
	jsr	DisasmLine
	lda	#145 ;Cursor UP!
	jsr	$ffd2

	jmp	DisasmAddress
;**************************************************************************
ScrollDisasmDown:
	lda	#$f0
	jsr	SubFromMemPtr
	lda	#$3d
	sta	$0326	;Stng av print
sdad_lp1:
	lda	$c1
	pha
	lda	$c2
	pha
	jsr	DisasmLine
	pla
	tay
	pla
	tax
	lda	$c4
	cmp	$c2
	bcc	sdad_skp2
	bne	sdad_skp1
	lda	$c1
	cmp	$c3
	bcs	sdad_skp2
sdad_skp1:
	jmp	sdad_lp1
sdad_skp2:
	stx	$c1
	sty	$c2
	lda	#$ca
	sta	$0326
	dec	$d6
	jmp	DisasmAddress
;**************************************************************************
AddToMemPtr:
	clc
	adc	$c3
	sta	$c1
	lda	$c4
	adc	#$00
	sta	$c2
	rts
SubFromMemPtr:
	clc
	adc	$c3
	sta	$c1
	lda	$c4
	sbc	#$00
	sta	$c2
	lda	#$ff
	sta	$d6
	rts
;**************************************************************************


;**************************************************************************
;**
;** Command 'A' - Assemble instruction.
;**
;******
Command_A:			;Assemble!
; get address
	jsr	GetWordAndSkip
; get mnemonic
	ldx	#0
cma_lp1:
	jsr	GetAsciiChar
	sta	$0204,x
	inx
	cpx	#3
	bne	cma_lp1
	
; preserve position
	tya
	pha
; get the rest of the line
cma_lp2:
	jsr	GetAsciiChar
	sta	$0204,x
	inx
	cpy	#39
	bcc	cma_lp2
; null terminate
	lda	#0
	sta	$0204-1,x
; restore position
	pla
	tay

; scan forward until the first hex char, if any.
cma_lp3:
	jsr	GetAsciiChar
	cpy	#$29
	bcc	cma_skp1
cma_ex1:
	jmp	CRMainLoopBig
cma_skp1:
	cmp	#$30
	bcs	cma_skp2
	cmp	#$07
	bcs	cma_lp3
cma_skp2:
	dey

; Try to get parameters
	jsr	GetHexByte
	sta	$c1
	jsr	GetHexByte
	sta	$c2

; flag as unknown
	lda	#$ff
	sta	$3a
; try to determine addressing mode and opcode
	jsr	AsmLine
; check if the instruction was found
	lda	$3a
	bne	cma_ex1		; no, exit.

; check addressing mode
	lda	$39
	bne	cma_skp3	; not relative

; relative branch
; convert to a one byte value in $c1, and flag as AFLG_2HEX.
	lda	$c2
	sec
	sbc	$c3
	sec
	sbc	#$02
	sta	$c1
	lda	#AFLG_2HEX
	sta	$39

cma_skp3:
; non relative addressing

	lda	$c2
	sta	$0204
	lda	$c1
	sta	$0205
	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2

	ldy	#0

; Number of bytes to X
	lda	$39
	and	#$f0
	tax

; put instruction byte
	lda	$0203
	jsr	PutByte_c3

; check the number of bytes to store
	cpx	#AFLG_NOHEX
	beq	cma_skp5
	cpx	#AFLG_2HEX
	bne	cma_skp4

; Put one byte to memory
	lda	$0205
	iny
	jsr	PutByte_c3

cma_skp4:
	cpx	#AFLG_4HEX
	bne	cma_skp5

; Put two bytes to memory  (LSB, MSB)
	lda	$0204
	iny
	jsr	PutByte_c3
	lda	$0205
	iny
	jsr	PutByte_c3

; All done, disassemble the line and print the next.
cma_skp5:
	lda	#"A"
	jsr	DisasmLine

	lda	#13
	jsr	$ffd2
	lda	#"A"
	jsr	$ffd2		;print 'Axxxx '
	lda	$c2
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte
	lda	#" "
	jsr	$ffd2
	jmp	MainLoopSmall
;**************************************************************************



;**************************************************************************
;**
;** Command 'D' - Disassemble memory.
;**
;******
Command_D:			;DISASSEMBLE!!!
	jsr	GetTwoHexWords
	cpy	#$26
	bcc	cmd_skp1
DisasmAddress:
	lda	$c1
	sta	$c3
	lda	$c2
	sta	$c4
cmd_skp1:
cmd_lp1:
	lda	#13
	jsr	$ffd2
	lda	#","
	jsr	DisasmLine
	ldy	#$00
	sty	$d3
	lda	$c4
	cmp	$c2
	bcc	cmd_ex1
	bne	cmd_skp3
	lda	$c3
	cmp	$c1
	bcs	cmd_skp3
cmd_ex1:
	jmp	MainLoopSmall
;***
cmd_skp3:
	lda	$91
	bmi	cmd_skp4
	bpl	cmd_ex1		;Break
cmd_skp4:
	jmp	cmd_lp1
;**************************************************************************


;**************************************************************************
MnemonicTable:
	dc.b	"A"&$1f,"D"&$1f,"C"&$1f,$61	;ADC
	dc.b	"N"&$1f|$40,"D"&$1f,$21		;AND
	dc.b	"S"&$1f|$40,"L"&$1f,$02		;ASL

	dc.b	"B"&$1f,"C"&$1f,"C"&$1f,$90	;BCC
	dc.b	"S"&$1f|$80,$b0			;BCS
	dc.b	"E"&$1f|$40,"Q"&$1f,$f0		;BEQ
	dc.b	"I"&$1f|$40,"T"&$1f,$24		;BIT
	dc.b	"M"&$1f|$40,"I"&$1f,$30		;BMI
	dc.b	"N"&$1f|$40,"E"&$1f,$D0		;BNE
	dc.b	"P"&$1f|$40,"L"&$1f,$10		;BPL
	ifconst	HAVE_DTV_V2
	dc.b	"R"&$1f|$40,"A"&$1f,$12		;BRA	; not complete
	dc.b	"K"&$1f|$80,$00			;BRK
	else
	dc.b	"R"&$1f|$40,"K"&$1f,$00		;BRK
	endif
	dc.b	"V"&$1f|$40,"C"&$1f,$50		;BVC
	dc.b	"S"&$1f|$80,$70			;BVS

	dc.b	"C"&$1f,"L"&$1f,"C"&$1f,$18	;CLC
	dc.b	"D"&$1f|$80,$d8			;CLD
	dc.b	"I"&$1f|$80,$58			;CLI
	dc.b	"V"&$1f|$80,$b8			;CLV
	dc.b	"M"&$1f|$40,"P"&$1f,$c1		;CMP
	dc.b	"P"&$1f|$40,"X"&$1f,$e0		;CPX
	dc.b	"Y"&$1f|$80,$c0			;CPY

	dc.b	"D"&$1f,"E"&$1f,"C"&$1f,$c2	;DEC
	dc.b	"X"&$1f|$80,$ca			;DEX
	dc.b	"Y"&$1f|$80,$88			;DEY
	
	dc.b	"E"&$1f,"O"&$1f,"R"&$1f,$41	;EOR

	dc.b	"I"&$1f,"N"&$1f,"C"&$1f,$e2	;INC
	dc.b	"X"&$1f|$80,$e8			;INX
	dc.b	"Y"&$1f|$80,$c8			;INY

	dc.b	"J"&$1f,"M"&$1f,"P"&$1f,$4c	;JMP
	dc.b	"S"&$1f|$40,"R"&$1f,$20		;JSR

	dc.b	"L"&$1f,"D"&$1f,"A"&$1f,$a1	;LDA
	dc.b	"X"&$1f|$80,$a2			;LDX
	dc.b	"Y"&$1f|$80,$a0			;LDY
	dc.b	"S"&$1f|$40,"R"&$1f,$42		;LSR

	dc.b	"N"&$1f,"O"&$1f,"P"&$1f,$ea	;NOP

	dc.b	"O"&$1f,"R"&$1f,"A"&$1f,$01	;ORA

	dc.b	"P"&$1f,"H"&$1f,"A"&$1f,$48	;PHA
	dc.b	"P"&$1f|$80,$08			;PHP
	dc.b	"L"&$1f|$40,"A"&$1f,$68		;PLA
	dc.b	"P"&$1f|$80,$28			;PLP

	dc.b	"R"&$1f,"O"&$1f,"L"&$1f,$22	;ROL
	dc.b	"R"&$1f|$80,$62			;ROR
	dc.b	"T"&$1f|$40,"I"&$1f,$40		;RTI
	dc.b	"S"&$1f|$80,$60			;RTS

	ifconst	HAVE_DTV_V2
	dc.b	"S"&$1f,"A"&$1f,"C"&$1f,$32-8	;SAC	; not complete
	dc.b	"I"&$1f|$40,"R"&$1f,$42-8	;SIR	; not complete
	dc.b	"B"&$1f|$40,"C"&$1f,$e1		;SBC
	else
	dc.b	"S"&$1f,"B"&$1f,"C"&$1f,$e1	;SBC
	endif
	dc.b	"E"&$1f|$40,"C"&$1f,$38		;SEC
	dc.b	"D"&$1f|$80,$f8			;SED
	dc.b	"I"&$1f|$80,$78			;SEI
	dc.b	"T"&$1f|$40,"A"&$1f,$81		;STA
	dc.b	"X"&$1f|$80,$82			;STX
	dc.b	"Y"&$1f|$80,$80			;STY

	dc.b	"T"&$1f,"A"&$1f,"X"&$1f,$aa	;TAX
	dc.b	"Y"&$1f|$80,$a8			;TAY
	dc.b	"S"&$1f|$40,"X"&$1f,$ba		;TSX
	dc.b	"X"&$1f|$40,"A"&$1f,$8a		;TXA
	dc.b	"S"&$1f|$80,$9a			;TXS
	dc.b	"Y"&$1f|$40,"A"&$1f,$98		;TYA
MnemonicTable_end:
;**************************************************************************


;**************************************************************************
;**
;** Place Mnemonic and Opcode in buffer
;**
;******
GetMnemonic:
	ldx	#0

	lda	MnemonicTable,y
	cmp	#$40	;Two Byte
	bcc	gmne_skp1
	inx		;Increase bytecounter
gmne_skp1:
	cmp	#$80	;One Byte
	bcc	gmne_skp2
	inx		;Increase ByteCounter
gmne_skp2:
	and	#$1f	;filterchar
gmne_lp1:
	sta	$0200,x
	iny
	lda	MnemonicTable,y
	inx
	cpx	#4
	bne	gmne_lp1
	cpy	#MnemonicTable_end-MnemonicTable+1 ;End of the list?
	rts
;**************************************************************************

AFLG_HASH	equ	%00000001
AFLG_X		equ	%00000010
AFLG_Y		equ	%00000100
AFLG_PAREN	equ	%00001000
AFLG_NOHEX	equ	%00010000
AFLG_2HEX	equ	%00100000
AFLG_4HEX	equ	%01000000
AFLG_MANYHEX	equ	%10000000
AFLG_NONE	equ	%00000000

	
DFLG_IMMEDIATE	equ	%00000001
DFLG_INDEX_X	equ	%00000010
DFLG_INDEX_Y	equ	%00000100
DFLG_INDIRECT	equ	%00001000
DFLG_1BYTE	equ	%00010000
DFLG_2BYTES	equ	%00100000
DFLG_3BYTES	equ	%01000000
DFLG_INVALID	equ	%11111111
DFLG_RELATIVE	equ	%00000000

	
;**************************************************************************
asml_Patch:
	ldx	#0
	ldy	#0
alph_lp1:
	lda	$0207,x
	beq	alph_skp3
	cmp	#"0"&$3f
	bcs	alph_skp1
	cmp	#("F"&$3f)+1
	bcs	alph_skp2
alph_skp1:
	iny
alph_skp2:
	inx
	bne	alph_lp1	; always taken
alph_skp3:
	lda	#AFLG_NOHEX
	cpy	#0
	beq	alph_ex1
	lda	#AFLG_2HEX
	cpy	#2
	beq	alph_ex1
	lda	#AFLG_4HEX
	cpy	#4
	beq	alph_ex1
	lda	#AFLG_MANYHEX
alph_ex1:
	ora	$39
	sta	$39
	rts
;***
asml_Patch2:
	lda	$0203
	cmp	#$a2		; LDX in MnemonicTable
	bne	alpt2_ex1
	lda	#$c0
alpt2_ex1:
	rts
;***
asml_ex2:
	jmp	CRMainLoopBig
;**************************************************************************


;**************************************************************************
AsmLine:
	ldy	#0
	sty	$39
asml_lp1:
	jsr	GetMnemonic
	bcs	asml_ex2
	ldx	#0
asml_lp2:
	lda	$0200,x
	cmp	$0204,x
	bne	asml_lp1
	inx
	cpx	#3
	bne	asml_lp2

	ldx	#0
asml_lp3:
	lda	$0207,x
	beq	asml_skp6	; End of line, skip out.
	cmp	#"#"&$3f
	bne	asml_skp1

	lda	#AFLG_HASH
	bne	asml_skp5	; always taken.

asml_skp1:
	cmp	#"X"&$3f
	bne	asml_skp2
	lda	#AFLG_X
	bne	asml_skp5	; always taken.

asml_skp2:
	cmp	#"Y"&$3f
	bne	asml_skp3
	lda	#AFLG_Y
	bne	asml_skp5	; always taken.

asml_skp3:
	cmp	#"("&$3f
	bne	asml_skp4
	lda	#AFLG_PAREN
	bne	asml_skp5	; always taken.

asml_skp4:
	lda	#AFLG_NONE
asml_skp5:
	ora	$39
	sta	$39
	inx
	bne	asml_lp3	; always taken

; line scanned for adressing modes, continue
asml_skp6:
; scan line for hex digits
	jsr	asml_Patch

; all done, now analyze

;***
; check for relative branches
;
	lda	$39
	cmp	#AFLG_4HEX
	bne	asml_CheckImplied

; check if this is a relative branch.
	lda	$0203
	ifconst HAVE_DTV_V2
	cmp	#$12		; =BRA
	beq	asml_skp7
	endif
	and	#$1f
	cmp	#$10
	bne	asml_CheckImplied	; no, continue check
asml_skp7:
	jmp	asml_ExitBranchOk	; yes, branch out.

;***
; Check for implied instructions
asml_CheckImplied:
	lda	$39
	cmp	#AFLG_NOHEX
	bne	asml_CheckImmediate

	lda	$0203
	bmi	asml_ExitOk_tmp1

	and	#$1f
	cmp	#$02
	bne	asml_ExitOk_tmp1
asml_ExitAddOffset08:
	lda	#$08
asml_ExitAddOffset_tmp1:
	jmp	asml_ExitAddOffset
asml_ExitOk_tmp1:
	jmp	asml_ExitOk

;***
; Check for Immediate instructions
;
asml_CheckImmediate:
	cmp	#AFLG_2HEX | AFLG_HASH
	bne	asml_CheckZpAbs

	jsr	asml_Patch2

	cmp	#$c0
	beq	asml_ExitOk_tmp1
	cmp	#$e0
	beq	asml_ExitOk_tmp1
	cmp	#$a0
	beq	asml_ExitOk_tmp1
	bne	asml_ExitAddOffset08	; always taken.

;***
; Check for zero page absolute instructions
;
asml_CheckZpAbs:
	cmp	#AFLG_2HEX
	bne	asml_CheckZpAbsX
	lda	$0203
	cmp	#$24
	beq	asml_ExitOk_tmp1
	lda	#$04
	bne	asml_ExitAddOffset_tmp1	; always taken.

;***
; Check for zero page absolute,X instructions
;
asml_CheckZpAbsX:
	cmp	#AFLG_2HEX | AFLG_X
	bne	asml_CheckZpAbsY
asml_ExitAddOffset14:
	lda	#$14
	bne	asml_ExitAddOffset_tmp1	; always taken.

;***
; Check for zero page absolute,Y instructions
;
asml_CheckZpAbsY:
	cmp	#AFLG_2HEX | AFLG_Y
	beq	asml_ExitAddOffset14
;***
; Check for absolute instructions
;
	cmp	#AFLG_4HEX
	bne	asml_CheckAbsX

	lda	$0203
	cmp	#$20
	beq	asml_ExitOk_tmp1
	cmp	#$4c
	beq	asml_ExitOk_tmp1
	cmp	#$24
	beq	asml_ExitAddOffset08
	lda	#$0c
	bne	asml_ExitAddOffset_tmp1	; always taken.

;***
; Check for absolute,X instructions
;
asml_CheckAbsX:
	cmp	#AFLG_4HEX | AFLG_X
	bne	asml_CheckAbsY
	lda	$0203
	cmp	#$80
	bne	asml_skp16
asml_ExitError:
	rts
asml_skp16:
	cmp	#$82
	beq	asml_ExitError
asml_ExitAddOffset1c:
	lda	#$1c
	bne	asml_ExitAddOffset_tmp1	; always taken.

;***
; Check for absolute,Y instructions
;
asml_CheckAbsY:
	cmp	#AFLG_4HEX | AFLG_Y
	bne	asml_CheckZpIndir
	lda	$0203
	cmp	#$a2
	beq	asml_ExitAddOffset1c
	lda	#$18
	bne	asml_ExitAddOffset_tmp1	; always taken.
;***
; Check for match on line 01, (zp),y instructions
;
asml_CheckZpIndir:
	lda	$0203
	and	#$1f
	cmp	#$01
	bne	asml_CheckIndir
;***
; Check for (zp,x) instructions
;
	lda	$39
	cmp	#AFLG_2HEX | AFLG_PAREN | AFLG_X
	beq	asml_ExitOk_tmp1
;***
; Check for (zp),y instructions
;
	cmp	#AFLG_2HEX | AFLG_PAREN | AFLG_Y
	bne	asml_ExitError
	lda	#$10
asml_ExitAddOffset_tmp2:
	jmp	asml_ExitAddOffset

;***
; Check for (indir) instructions
; JMP (abs)
;
asml_CheckIndir:
	lda	$39
	cmp	#AFLG_4HEX | AFLG_PAREN
	bne	asml_ExitError

	lda	$0203
	cmp	#$4c
	bne	asml_ExitError

	lda	#$20
	bne	asml_ExitAddOffset_tmp2	; always taken.

;***
; Do not add offset and exit.
;
asml_ExitBranchOk:
	lda	#AFLG_NONE	; flag for relative branch
	sta	$39
	beq	asml_ExitOk	; always taken.

;***
; Add offset and exit.
;
asml_ExitAddOffset:
	clc
	adc	$0203
	sta	$0203
asml_ExitOk:
; flag ok.
	inc	$3a
	rts
;**************************************************************************


;**************************************************************************
;**
;** Determine opcode length...
;**
;******
TestOpcode:
;***
; Step one, fetch opcode flags
;

; Major addressing mode -> X
	lda	$0204
	and	#$1f
	tax

; Check for exceptions
	ldy	#DisAsmTable3_end-DisAsmTable3-1
to_lp1:
	lda	DisAsmTable3,y
	cmp	$0204
	beq	to_skp1		
	dey
	bpl	to_lp1

; If no exceptions, load from DisAsmTable1  
	lda	DisAsmTable1,x
to_lp2:
	sta	$39
	cmp	#DFLG_INVALID	;Unknown opcode!
	bne	to_skp2		;No, skip to next...
	rts
;***
; if exceptions, load from exception table
to_skp1:
	lda	DisAsmTable4,y
	bcs	to_lp2		; always taken.


;***
; Step two, fetch opcode
;
to_skp2:
	ldy	#DisAsmTable5_end-DisAsmTable5-1
to_lp3:
	lda	DisAsmTable5,y
	cmp	$0204
	beq	to_skp3
	dey
	bpl	to_lp3

	lda	$0204
	sec
	sbc	DisAsmTable2,x
	sta	$0204
	bcs	to_skp4		; always taken.

to_skp3:
	lda	DisAsmTable6,y
	sta	$0204

to_skp4:

;***
; Step three, fetch mnemonic corresponding to opcode
;
	ldy	#0
to_lp4:
	jsr	GetMnemonic
	bcs	to_ex1	;Slut p listan
	lda	$0203
	cmp	$0204	;Rtt opkod?
	bne	to_lp4	;Nej... testa nsta!
to_ex1:
	inc	$3a
	rts
;**************************************************************************


;**************************************************************************
;**
;** Disassemble a byte sequence
;**
;******
DisasmLine:
	pha

	ldy	#0
	jsr	GetByte_c1
	sta	$0204
	lda	#$ff
	sta	$3a

	jsr	TestOpcode

	lda	$3a
	beq	dal_skp1

	lda	#$ff
	sta	$0200
	sta	$0201	;Unknown opcode!
	sta	$0202
	lda	#DFLG_1BYTE
	sta	$39

dal_skp1:
	lda	#0	;Column 0
	sta	$d3	
	pla
	jsr	$ffd2	;Show header char.

	ldy	#1
	lda	#" "
dal_lp2:
	sta	($d1),y	;Clear row!!!
	iny
	cpy	#40
	bne	dal_lp2

	lda	$c2
	jsr	PrintHexByte	;Write address...
	lda	$c1
	jsr	PrintHexByte

	lda	$39		; flags
	and	#%11110000
	tax
	bne	dal_skp2
	ldx	#DFLG_2BYTES ; no flags (branch), just set 2 bytes
dal_skp2:

	ldy	#0
	jsr	GetByte_c1
	jsr	PrintSpaceHexByte	;Skriv ut instruktions bytarna
	cpx	#DFLG_1BYTE
	beq	dal_skp3
	iny
	jsr	GetByte_c1
	sta	$0204
	jsr	PrintSpaceHexByte	;(1-3 st)
	cpx	#DFLG_2BYTES
	beq	dal_skp3
	iny
	jsr	GetByte_c1
	sta	$0205
	jsr	PrintSpaceHexByte	;beroende p $39
dal_skp3:

	tya
	sec
	adc	$c1		;Addera instruktionslngd p $c1/$c2
	sta	$c1
	bcc	dal_skp4
	inc	$c2
dal_skp4:

	lda	#15		;Kolumn 15
	sta	$d3
	ldy	#0
	jmp	PrintInstruction
;**************************************************************************

;                6510 Instructions by Addressing Modes

;off- ++++++++++ Positive ++++++++++  ---------- Negative ----------
;set  00      20      40      60      80      a0      c0      e0     mode
;
;+00  BRK     JSR     RTI     RTS     NOP*    LDY     CPY     CPX    Impl/immed
;+01  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    (indir,x)
;+02   t       t      SIR      t      NOP*t   LDX     NOP*t   NOP*t   ?/immed
;+03  SLO*    RLA*    SRE*    RRA*    SAX*    LAX*    DCP*    ISB*   (indir,x)
;+04  NOP*    BIT     NOP*    NOP*    STY     LDY     CPY     CPX    Zeropage
;+05  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    Zeropage
;+06  ASL     ROL     LSR     ROR     STX     LDX     DEC     INC    Zeropage
;+07  SLO*    RLA*    SRE*    RRA*    SAX*    LAX*    DCP*    ISB*   Zeropage
;
;+08  PHP     PLP     PHA     PLA     DEY     TAY     INY     INX    Implied
;+09  ORA     AND     EOR     ADC     NOP*    LDA     CMP     SBC    Immediate
;+0a  ASL     ROL     LSR     ROR     TXA     TAX     DEX     NOP    Accu/impl
;+0b  ANC**   ANC**   ASR**   ARR**   ANE**   LXA**   SBX**   SBC*   Immediate
;+0c  NOP*    BIT     JMP     JMP  () STY     LDY     CPY     CPX    Absolute
;+0d  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    Absolute
;+0e  ASL     ROL     LSR     ROR     STX     LDX     DEC     INC    Absolute
;+0f  SLO*    RLA*    SRE*    RRA*    SAX*    LAX*    DCP*    ISB*   Absolute
;
;+10  BPL     BMI     BVC     BVS     BCC     BCS     BNE     BEQ    Relative
;+11  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    (indir),y
;+12  BRA     SAC      t       t       t       t       t       t      ?/DTV
;+13  SLO*    RLA*    SRE*    RRA*    SHA**   LAX*    DCP*    ISB*   (indir),y
;+14  NOP*    NOP*    NOP*    NOP*    STY     LDY     NOP*    NOP*   Zeropage,x
;+15  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    Zeropage,x
;+16  ASL     ROL     LSR     ROR     STX  y) LDX  y) DEC     INC    Zeropage,x
;+17  SLO*    RLA*    SRE*    RRA*    SAX* y) LAX* y) DCP*    ISB*   Zeropage,x
;
;+18  CLC     SEC     CLI     SEI     TYA     CLV     CLD     SED    Implied
;+19  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    Absolute,y
;+1a  NOP*    NOP*    NOP*    NOP*    TXS     TSX     NOP*    NOP*   Implied
;+1b  SLO*    RLA*    SRE*    RRA*    SHS**   LAS**   DCP*    ISB*   Absolute,y
;+1c  NOP*    NOP*    NOP*    NOP*    SHY**   LDY     NOP*    NOP*   Absolute,x
;+1d  ORA     AND     EOR     ADC     STA     LDA     CMP     SBC    Absolute,x
;+1e  ASL     ROL     LSR     ROR     SHX**y) LDX  y) DEC     INC    Absolute,x
;+1f  SLO*    RLA*    SRE*    RRA*    SHA**y) LAX* y) DCP*    ISB*   Absolute,x
;
;
;        ROR intruction is available on MC650x microprocessors after
;        June, 1976.
;
;
;        Legend:
;
;        t       Jams the machine
;        *t      Jams very rarely
;        *       Undocumented command
;        **      Unusual operation
;        y)      indexed using Y instead of X
;        ()      indirect instead of absolute
;
;        Note that the NOP instructions do have other addressing modes
;        than the implied addressing. The NOP instruction is just like
;        any other load instruction, except it does not store the
;        result anywhere nor affects the flags.
;


;**************************************************************************
;**
;** Disassembler tables!
;**
;******
; opcode flags for regulars
DisAsmTable1:
	dc.b	DFLG_2BYTES | DFLG_IMMEDIATE ; 00 Impl/immed
	dc.b	DFLG_2BYTES | DFLG_INDIRECT | DFLG_INDEX_X ; 01 (indir,x)
	dc.b	DFLG_INVALID	             ; 02 (? /immed) (undoc + DTV)
	dc.b	DFLG_INVALID	             ; 03 (indir,x)  (undoc)
	dc.b	DFLG_2BYTES	             ; 04 Zeropage
	dc.b	DFLG_2BYTES                  ; 05 Zeropage
	dc.b	DFLG_2BYTES                  ; 06 Zeropage 
	dc.b	DFLG_INVALID                 ; 07 Zeropage
	dc.b	DFLG_1BYTE                   ; 08 Implied 
	dc.b	DFLG_2BYTES | DFLG_IMMEDIATE ; 09 Immediate 
	dc.b	DFLG_1BYTE                   ; 0a Accu/impl 
	dc.b	DFLG_INVALID                 ; 0b Immediate  (undoc)
	dc.b	DFLG_3BYTES                  ; 0c Absolute
	dc.b	DFLG_3BYTES	             ; 0d Absolute
	dc.b	DFLG_3BYTES	             ; 0e Absolute
	dc.b	DFLG_INVALID	             ; 0f Absolute   (undoc)
	dc.b	DFLG_RELATIVE	             ; 10 Relative
	dc.b	DFLG_2BYTES | DFLG_INDIRECT | DFLG_INDEX_Y ; 11 (indir),y
	dc.b	DFLG_INVALID	             ; 12 ?  (undoc)
	dc.b	DFLG_INVALID	             ; 13 (indir),y  (undoc)
	dc.b	DFLG_INVALID	             ; 14 Zeropage,x (undoc)
	dc.b	DFLG_2BYTES | DFLG_INDEX_X   ; 15 Zeropage,x 
	dc.b	DFLG_2BYTES | DFLG_INDEX_X   ; 16 Zeropage,x
	dc.b	DFLG_INVALID	             ; 17 Zeropage,x (undoc)
	dc.b	DFLG_1BYTE 	             ; 18 Implied 
	dc.b	DFLG_3BYTES | DFLG_INDEX_Y   ; 19 Absolute,y 
	dc.b	DFLG_INVALID	             ; 1a Implied (undoc)
	dc.b	DFLG_INVALID	             ; 1b Absolute,y (undoc) 
	dc.b	DFLG_INVALID	             ; 1c Absolute,x (undoc)
	dc.b	DFLG_3BYTES | DFLG_INDEX_X   ; 1d Absolute,x 
	dc.b	DFLG_3BYTES | DFLG_INDEX_X   ; 1e Absolute,x 
	dc.b	DFLG_INVALID	             ; 1f Absolute,x (undoc) 
; opcode tag offsets for regulars
DisAsmTable2:
	dc.b	$00	; 00 Impl/immed
	dc.b	$00	; 01 (indir,x)
	dc.b	$00	; 02 (? /immed) (undoc + DTV)
	dc.b	$00	; 03 (indir,x)  (undoc)
	dc.b	$04	; 04 Zeropage
	dc.b	$04	; 05 Zeropage
	dc.b	$04	; 06 Zeropage
	dc.b	$00	; 07 Zeropage
	dc.b	$00     ; 08 Implied
	dc.b	$08     ; 09 Immediate
	dc.b	$00     ; 0a Accu/impl
	dc.b	$00     ; 0b Immediate  (undoc)
	dc.b	$0c     ; 0c Absolute
	dc.b	$0c     ; 0d Absolute
	dc.b	$0c     ; 0e Absolute
	dc.b	$00	; 0f Absolute   (undoc)
	dc.b	$00	; 10 Relative
	dc.b	$10	; 11 (indir),y
	dc.b	$00	; 12 ?  (undoc)
	dc.b	$00	; 13 (indir),y  (undoc)
	dc.b	$14	; 14 Zeropage,x (undoc)
	dc.b	$14	; 15 Zeropage,x
	dc.b	$14	; 16 Zeropage,x
	dc.b	$00	; 17 Zeropage,x (undoc)
	dc.b	$00	; 18 Implied 
	dc.b	$18	; 19 Absolute,y
	dc.b	$00	; 1a Implied	(undoc)
	dc.b	$00	; 1b Absolute,y (undoc)
	dc.b	$1c	; 1c Absolute,x (undoc)
	dc.b	$1c	; 1d Absolute,x
	dc.b	$1c	; 1e Absolute,x
	dc.b	$00	; 1f Absolute,x (undoc)
;**************************************************************************
; Table of exceptions
DisAsmTable3:
	dc.b	$00	;BRK
	dc.b	$20	;JSR $xxxx
	dc.b	$40	;RTI
	dc.b	$60	;RTS
	dc.b	$80	;???
	dc.b	$a2	;LDX #$yy
	dc.b	$04	;???
	dc.b	$44	;???
	dc.b	$64	;???
	dc.b	$89	;???
	dc.b	$6c	;JMP ($xxxx)
	dc.b	$0c	;???
	dc.b	$94	;STY $zz,X
	dc.b	$b4	;LDY $zz,X
	dc.b	$96	;STX $zz,Y
	dc.b	$b6	;LDX $zz,Y
	dc.b	$9a	;TXS
	dc.b	$ba	;TSX
	dc.b	$bc	;LDY $xxxx,X
	dc.b	$9e	;???
	dc.b	$be	;LDX $xxxx,Y
	ifconst	HAVE_DTV_V2
	dc.b	$12	;BRA $xxxx
	dc.b	$32	;SAC #$xx
	dc.b	$42	;SIR #$xx
	endif
DisAsmTable3_end:
; Flags for exceptions
DisAsmTable4:
	dc.b	DFLG_1BYTE	             ;BRK
	dc.b	DFLG_3BYTES	             ;JSR $xxxx
	dc.b	DFLG_1BYTE	             ;RTI
	dc.b	DFLG_1BYTE	             ;RTS
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_2BYTES | DFLG_IMMEDIATE ;LDX #$yy
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_3BYTES | DFLG_INDIRECT  ;JMP ($xxxx)
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_2BYTES | DFLG_INDEX_X   ;STY $zz,X
	dc.b	DFLG_2BYTES | DFLG_INDEX_X   ;LDY $zz,X
	dc.b	DFLG_2BYTES | DFLG_INDEX_Y   ;STX $zz,Y
	dc.b	DFLG_2BYTES | DFLG_INDEX_Y   ;LDX $zz,Y
	dc.b	DFLG_1BYTE	             ;TXS
	dc.b	DFLG_1BYTE	             ;TSX
	dc.b	DFLG_3BYTES | DFLG_INDEX_X   ;LDY $xxxx,X
	dc.b	DFLG_INVALID	             ;???
	dc.b	DFLG_3BYTES | DFLG_INDEX_Y   ;LDX $xxxx,Y
	ifconst	HAVE_DTV_V2
	dc.b	DFLG_RELATIVE		     ;BRA $xxxx
	dc.b	DFLG_2BYTES | DFLG_IMMEDIATE  ;SAC #$xx
	dc.b	DFLG_2BYTES | DFLG_IMMEDIATE  ;SIR #$xx
	endif
;**************************************************************************
; Table of opcode exceptions
DisAsmTable5:
	dc.b	$24	;BIT $zz
	dc.b	$0a	;ASL
	dc.b	$2a	;ROL
	dc.b	$4a	;LSR
	dc.b	$6a	;ROR
	dc.b	$2c	;BIT $xxxx
	dc.b	$4c	;JMP $xxxx
	dc.b	$6c	;JMP ($xxxx)
	ifconst	HAVE_DTV_V2
	dc.b	$32	;SAC #$xx
	dc.b	$42	;SIR #$xx
	endif
DisAsmTable5_end:
; Tags into the mnemonic table.
DisAsmTable6:
	dc.b	$24	;BIT $zz
	dc.b	$02	;ASL
	dc.b	$22	;ROL
	dc.b	$42	;LSR
	dc.b	$62	;ROR
	dc.b	$24	;BIT $xxxx
	dc.b	$4c	;JMP $xxxx
	dc.b	$4c	;JMP ($xxxx)
	ifconst	HAVE_DTV_V2
	dc.b	$32-8	;SAC #$xx
	dc.b	$42-8	;SIR #$xx
	endif
;**************************************************************************



;**************************************************************************
PrintInstruction:
pins_lp1:
	lda	$0200,y
	clc
	adc	#$40		;Skriv mnemonic
	jsr	$ffd2
	iny
	cpy	#$03
	bne	pins_lp1

	lda	$39
	bne	pins_skp3

;*** Rel address ***
	lda	$0204
	bmi	pins_skp1
	clc
	adc	$c1
	sta	$0204
	lda	$c2
	adc	#$00
	sta	$0205
	clc
	bcc	pins_skp2
pins_skp1:
	clc
	adc	$c1
	sta	$0204
	lda	$c2
	sbc	#$00
	sta	$0205
pins_skp2:
	lda	#DFLG_3BYTES
	sta	$39

;*** Skriv hexxx ***
pins_skp3:
	lda	#" "
	jsr	$ffd2

	lda	$39
	cmp	#DFLG_2BYTES | DFLG_IMMEDIATE
	bne	pins_skp4

;*** immediate ***
	lda	#"#"
	jsr	$ffd2

pins_skp4:
	lda	$39
	and	#DFLG_INDIRECT
	beq	pins_skp5

;*** Indirekt ***
	lda	#"("
	jsr	$ffd2

pins_skp5:
	lda	$39
	cmp	#DFLG_1BYTE
	beq	pins_ex1

	lda	#"$"
	jsr	$ffd2

	lda	$39
	and	#DFLG_3BYTES
	beq	pins_skp6

;*** Word ***
	lda	$0205
	jsr	PrintHexByte
;*** Byte ***
pins_skp6:
	lda	$0204
	jsr	PrintHexByte

	lda	$39
	and	#DFLG_INDEX_X
	beq	pins_skp7

;*** Index X ***
	lda	#","
	jsr	$ffd2
	lda	#"X"
	jsr	$ffd2

pins_skp7:
	lda	$39
	and	#DFLG_INDIRECT
	beq	pins_skp8

;*** Indirekt ***
	lda	#")"
	jsr	$ffd2

pins_skp8:
	lda	$39
	and	#DFLG_INDEX_Y
	beq	pins_ex1

;*** Index Y ***
	lda	#","
	jsr	$ffd2
	lda	#"Y"
	jsr	$ffd2

pins_ex1:
	rts
;**************************************************************************



;**************************************************************************
GetWordAndSkip:
	jsr	GetHexWord
SkipJunk:
	ldx	#$00
gwas_lp1:
	iny
	cpy	#$28
	bcs	gwas_ex1
	jsr	GetFromScreen
	bcc	gwas_skp1
	ldx	#$ff
gwas_skp1:
	inx
	cpx	#$03
	bne	gwas_lp1
	dey
	dey
	dey
gwas_ex1:
	rts
;**************************************************************************




;**************************************************************************
;**
;** Command ',' - Enter codedata
;**
;******
Command_Comma:			;Enter code data
	jsr	GetHexWord

	ldx	#$ff
	stx	$c1

	jsr	GetHexByte
	sta	$0210
	jsr	cmcm_Check
	bcc	cmcm_skp1

	jsr	GetHexByte
	sta	$0211
	inc	$c1
	jsr	cmcm_Check
	bcc	cmcm_skp1

	jsr	GetHexByte
	sta	$0212
	inc	$c1
cmcm_skp1:
	inc	$c1
	inc	$c1

	ldy	#$00
cmcm_lp2:
	lda	$0210,y
	jsr	PutByte_c3
	iny
	cpy	$c1
	bne	cmcm_lp2

	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2
	lda	#","
	jsr	DisasmLine

	lda	#13
	jsr	$ffd2
	lda	#","
	jsr	$ffd2
	lda	$c2		;Print Newline!
	jsr	PrintHexByte
	lda	$c1
	jsr	PrintHexByte
	lda	#" "
	jsr	$ffd2
	jmp	MainLoopSmall
;***
cmcm_Check:
	sty	$39
	cpy	#$28
	bcs	cmch_ex2
	jsr	GetAsciiChar
	sty	$3a
	ldy	$39
	jsr	SkipJunk
	cpy	#$27
	bcs	cmch_skp1
	cpy	$3a
cmch_ex3:
	ldy	$39
	rts
cmch_skp1:
	lda	$3a
	cmp	#$27
	bcs	cmch_ex1
	sec
	bcs	cmch_ex3
cmch_ex1:
	clc
	rts
cmch_ex2:
	jmp	CRMainLoopBig
;**************************************************************************


;**************************************************************************
;**
;** Command 'H' - Hunt memoryblock
;**
;******
Command_H:			;HUNT!!
	jsr	GetHexWord
	lda	$c3
	sta	$c1
	lda	$c4
	sta	$c2
	jsr	GetHexWord
	jsr	CheckSearchString

	ldx	#$ff
cmh_lp2:
	inx
	jsr	GetAsciiChar
	dey
	cmp	#$18
	bne	cmh_skp1
	lda	#$01
	bne	cmh_skp2
cmh_skp1:
	lda	#$ff
cmh_skp2:
	sta	$0220,x
	jsr	GetHexByte
	cpy	#$28
	bcs	cmh_skp3
	sta	$0210,x
	bcc	cmh_lp2
cmh_skp3:
	lda	#$00
	sta	$0220,x

cmh_StringEntry:
	lda	#13
	jsr	$ffd2
cmh_lp3:
	ldy	#0
cmh_lp4:
	jsr	GetByte_c1
	tax
	lda	$0220,y
	beq	cmh_skp5
	cmp	#$01
	beq	cmh_skp4
	txa
	cmp	$0210,y
	bne	cmh_skp6
cmh_skp4:
	iny
	bne	cmh_lp4
cmh_skp5:
	lda	$c2
	jsr	PrintSpaceHexByte
	lda	$c1
	jsr	PrintHexByte
cmh_skp6:
	inc	$c1
	bne	cmh_skp7
	inc	$c2
cmh_skp7:
	lda	$c1
	cmp	$c3
	lda	$c2
	sbc	$c4
	beq	cmh_skp8
	lda	$91
	bmi	cmh_lp3
cmh_skp8:
	lda	$d3
	beq	cmh_ex1
	jmp	CRMainLoopSmall
cmh_ex1:
	jmp	MainLoopSmall
;**************************************************************************
CheckSearchString:
	jsr	CheckEnd
	tya
	pha
	jsr	GetAsciiChar
	cmp	#34	; '"'
	beq	css_skp1
	pla
	tay
	rts
css_skp1:
	dey
	jsr	GetFileName
	ldy	#$0f
css_lp1:
	lda	$0201,y
	sta	$0210,y
	lda	#$ff
	sta	$0220,y
	dey
	bpl	css_lp1
	ldy	$b7
	lda	#$00
	sta	$0220,y
	jmp	cmh_StringEntry
;**************************************************************************


;**************************************************************************
;**
;** Command 'B' - Set breakpoint.
;**
;******
Command_B:			;Set breakpoint
	jsr	GetHexByte
	sta	BRK1Store+1
	jsr	GetHexByte
	sta	BRK1Store
	jmp	CRMainLoopBig
;**************************************************************************


;**************************************************************************
;*
;* Command 'N' - Relocate code or tables.
;*
;***
Command_N:			;Newlocate!
	jsr	GetTwoHexWords
	jsr	GetHexByte
	sta	$bf
	jsr	GetHexByte
	sta	$be
	jsr	GetHexByte
	sta	$b1
	jsr	GetHexByte
	sta	$b0
	jsr	GetHexByte
	sta	$b5
	jsr	GetHexByte
	sta	$b4
	cpy	#$28
	bcc	cmn_skp1
	jmp	CRMainLoopBig
cmn_skp1:

cmn_lp2:
	lda	#$ff
	sta	$3a
	ldy	#0
	jsr	GetByte_c1
	sta	$0204
	jsr	TestOpcode
	lda	$3a
	beq	cmn_skp2
	dec	$d6
	jmp	DisasmAddress
cmn_skp2:
	lda	$39
	bne	cmn_skp3
	lda	#$20
	sta	$39
cmn_skp3:
	and	#$40
	beq	cmn_skp4
	ldy	#$01
	jsr	GetByte_c1
	sta	$bb
	iny
	jsr	GetByte_c1
	sta	$bc
	lda	$bb
	cmp	$b0
	lda	$bc
	sbc	$b1
	bcc	cmn_skp4
	lda	$b4
	cmp	$bb
	lda	$b5
	sbc	$bc
	bcc	cmn_skp4
	lda	$bb
	clc
	adc	$be
	sta	$bb
	lda	$bc
	adc	$bf
	jsr	PutByte_c1
	dey
	lda	$bb
	jsr	PutByte_c1
cmn_skp4:
	lda	$39
	and	#$f0
	ldy	#$01
	cmp	#$10
	beq	cmn_skp5
	iny
	cmp	#$20
	beq	cmn_skp5
	iny
cmn_skp5:
	tya
	clc
	adc	$c1
	sta	$c1
	bcc	cmn_skp6
	inc	$c2
cmn_skp6:
	lda	$c4
	cmp	$c2
	bcc	cmn_ex1
	bne	cmn_skp7
	lda	$c3
	cmp	$c1
	bcs	cmn_skp7
cmn_ex1:
	jmp	CRMainLoopBig
cmn_skp7:
	lda	$91		;TestBreak
	bpl	cmn_ex1
	jmp	cmn_lp2
;**************************************************************************




;**************************************************************************
BRKNMIEntry:
	pla
	tay
	pla
	tax
	pla
BRKNMIEntry2:
	sei
	sta	ACStore
	stx	XRStore
	sty	YRStore
	pla
	sta	SRStore
	cld
	pla
	clc
	adc	#$ff
	sta	PCStore
	pla
	adc	#$ff
	sta	PCStore+1
	tsx
	stx	SPStore

	lda	$01
	sta	MemoryStore
	lda	#$07
	sta	$01

	lda	rh_sm1+1	;break or nmi?
	cmp	#<BRKNMIEntry2
	lda	#<BRKNMIEntry
	sta	rh_sm1+1
	bcs	bne_skp1	;nmi, skip to nmi routine.

; break, just dump registers.
	lda	#13
	jsr	$ffd2
	lda	#"B"
DumpRegsGoMain:
	jsr	$ffd2
	lda	#"*"
	jsr	$ffd2
	jmp	Command_R
;**************************************************************************
bne_skp1:
	inc	PCStore
	bne	bne_skp2
	inc	PCStore+1
bne_skp2:
	lda	$dd0d
	and	#$01
	bne	bne_skp3
	jmp	RestoreEntry
bne_skp3:
	cli
	lda	ModeStore
	bne	bne_skp4
	jmp	RunNextInstruction
;***
bne_skp4:
	bmi	bne_skp5

	lda	PCStore		;Check if breakpoint was reached!
	cmp	BRK1Store
	bne	bne_skp6
	lda	PCStore+1
	cmp	BRK1Store+1
	bne	bne_skp6

bne_skp5:
	lda	#13
	jsr	$ffd2
	jsr	cmr_DumpRegs
	lda	PCStore
	sta	$c1
	lda	PCStore+1
	sta	$c2
	lda	#$2c
	jsr	DisasmLine
bne_lp1:
	jsr	$ffe4
	beq	bne_lp1
	cmp	#"J"
	bne	bne_skp6
	lda	#$00
	sta	ModeStore
bne_skp6:
	lda	$91	;BreakCheck
	bmi	DoWalk
	jmp	CRMainLoopBig
;**************************************************************************



;**************************************************************************
DoWalk:
	ldx	#dw_WalkerLen-1
dw_lp1:
	lda	dw_WalkerSt,x
	sta	dw_Walker,x
	dex
	bpl	dw_lp1
	sei
dw_skp1:
	lda	#$7f
	sta	$dd0d
	lda	$dd0d
	lda	#$81
	sta	$dd0d
	lda	#$00
	sta	$dd05
	lda	#17
	jsr	SetTimer
	ldx	SPStore
	txs
	lda	SRStore
	pha
	ldx	XRStore
	ldy	YRStore
	jmp	dw_Walker
;***
dw_WalkerSt:
	rorg	$0200
dw_Walker:
	lda	#$19
	sta	$dd0e
dw_Walker2:
	lda	MemoryStore
	sta	$01
	lda	ACStore
	plp
	jmp	(PCStore)
	rend
dw_WalkerLen	equ . - dw_WalkerSt


;*** Get PC... maybe!!! ***
GetPC:
	jsr	GetHexByte
	cpy	#$28
	bcs	gp_ex1
	sta	PCStore+1
	jsr	GetHexByte
	sta	PCStore
gp_ex1:
	rts
;**************************************************************************



;**************************************************************************
;**
;** Command 'W' - Walk code
;**
;******
Command_W:				;Walk
	lda	#$80
cmw_skp1:
	sta	ModeStore
	jsr	GetPC
	clc
	bcc	DoWalk
;**************************************************************************


;**************************************************************************
;**
;** Command 'Q' - Quicktrace code.
;**
;******
Command_Q:				;Quicktrace
	lda	#$40
	bne	cmw_skp1
;**************************************************************************

;**************************************************************************
RunNextInstruction:
	sei
	pla
	clc
	adc	#$01
	sta	$c1
	pla
	adc	#$00
	sta	$c2
	lda	#>(dw_Walker3-1)
	pha
	lda	#<(dw_Walker3-1)
	pha
	ldx	#0
rni_lp1:
	lda	dw_Walker3St,x
	sta	dw_Walker3,x
	lda	dw_Walker2,x
	sta	$0200,x
	inx
	cpx	#$10
	bne	rni_lp1
	beq	dw_skp1
;***

dw_Walker3St:
	rorg	$0110
dw_Walker3:
	php
	pha
	lda	$01
	sta	MemoryStore
	lda	#$07
	sta	$01
	jmp	FastNMIEntry
	rend
dw_Walker3Len	equ	. - dw_Walker3St
	
;**************************************************************************
FastNMIEntry:
	sei
	pla
	sta	ACStore
	pla
	sta	SRStore
	stx	XRStore
	sty	YRStore
	cld
	lda	#$80
	sta	ModeStore
	lda	$c1
	sta	PCStore
	lda	$c2
	sta	PCStore+1
	jmp	bne_skp5
;**************************************************************************


;**************************************************************************
;**
;** Command 'E' - Install breakhandler
;**
;******
Command_E:
	ldx	#cme_BreakHandlerLen-1
cme_lp1:
	lda	cme_BreakHandlerSt,x
	sta	cme_BreakHandler,x
	dex
	bpl	cme_lp1
	jmp	CRMainLoopBig
;***
cme_BreakHandlerSt	equ	.
	rorg	$ffe3
cme_BreakHandler:
cebh_IRQ:
	pha
	txa
	pha
	tya
	pha
	tsx
	lda	$0104,x
	and	#$10
	beq	cebh_skp1
	jmp	($0316)
cebh_skp1:
	jmp	($fffc)
cebh_NMI:
	sei
	jmp	($0318)
	dc.w	cebh_NMI,0,cebh_IRQ
	rend
cme_BreakHandlerLen	equ	.-cme_BreakHandlerSt
;**************************************************************************



;**************************************************************************
XLateChar:
	cmp	#$5e
	bne	xlc_skp1
	lda	#$de
	inc	$d3
	bne	xlc_ex1
xlc_skp1:
	jsr	GetCharFromScreen
xlc_ex1:
	rts
;**************************************************************************


;**************************************************************************
GetFromScreen:
	lda	($d1),y
	cmp	#$3f
	beq	gfs_ex1
	cmp	#$20
	rts
gfs_ex1:
	clc
	rts
;**************************************************************************


InstallLoadSavePatch:
	ldx	#0
ilsp_lp1:
	lda	ilsp_Handler_st,x
	sta	ilsp_Handler,x
	inx
	cpx	#ilsp_Handler_end-ilsp_Handler_st
	bne	ilsp_lp1

; set up routine with the current banking options
	sir	$d2
	sty	ilsph_sm1+1
	sir	$12

	lda	#<LoadPatch_ram
	sta	$0330
	lda	#>LoadPatch_ram
	sta	$0331		;LoadVector

	lda	#<SavePatch_ram
	sta	$0332
	lda	#>SavePatch_ram
	sta	$0333		;SaveVector
	rts

ilsp_Handler_st:
	rorg	EndBreakHandler
ilsp_Handler:
ilsph_Preserve:
	sir	$d8
	sty	ilsph_sm3+1
	stx	ilsph_sm4+1
ilsph_sm1:
	ldy	#$06
	ldx	#DEFAULT_REG8
	sir	$12
	rts
LoadPatch_ram:
	jsr	ilsph_Preserve
	jsr	LoadPatch
	bra	ilsph_Restore
SavePatch_ram:
	jsr	ilsph_Preserve
	jsr	SavePatch
ilsph_Restore:
	sir	$d8
ilsph_sm3:
	ldy	#$01
ilsph_sm4:
	ldx	#%01010101
	sir	$12
	rts
ilsp_End:
	rend
ilsp_Handler_end:
	echo	"[monitor] load/save handler:",ilsp_Handler,"-",ilsp_End

;**************************************************************************
;**
;** Kernal LOAD patch
;**
;******
LoadPatch:
	pha
	lda	BankStore
	sta	$a7
	lda	#$00
	sta	BankStore
	pla
	jsr	clone_f4a5
	pha
	lda	$a7
	sta	BankStore
	pla
	rts
	
clone_ffd5:	
	stx	$c3
	sty	$c4
clone_f4a5:
	sta	$93
	tay
	beq	lp_skp1
	jmp	($fd4c)		; verify
lp_skp1:
	ldy	$ba
	cpy	#1
	bne	lp_skp2
; Flash load
; the original load vector from the table
	jmp	($fd4c) 
lp_skp2:
	jsr	NormalDiskLoad		;NormalDiskLoad
	php
	jsr	re_InstallBRKNMI
	plp
	jmp	$f5aa
;**************************************************************************


;**************************************************************************
;**
;** Kernal SAVE patch
;**
;******
SavePatch:
	pha
	lda	BankStore
	sta	$a7
	lda	RomStore
	sta	$a8
	lda	#$00
	sta	BankStore
	lda	RomStore
	ora	#%00010000
	sta	RomStore
	pla
	jsr	clone_f5ed
	pha
	lda	$a8
	sta	RomStore
	lda	$a7
	sta	BankStore
	pla
	rts

	
clone_ffd8:
	stx	$ae
	sty	$af
	tax
	lda	$00,x
	sta	$c1
	lda	$01,x
	sta	$c2
clone_f5ed:
	jsr	$fb8e		; set $ac/$ad from $c1/$c2
clone2_f5ed:
	ldy	$ba
	cpy	#1
	bne	sp_skp1
	jmp	$f65f		;NormalTapeSave
sp_skp1:
	jsr	NormalDiskSave		;NormalDiskSave
	jsr	re_InstallBRKNMI
	jmp	$f5a9
;**************************************************************************


;**************************************************************************
NormalDiskLoad:
	tya
	pha
	ldy	#$00
	lda	($bb),y
	cmp	#"$"
	bne	ndl_skp1
	pla
	tay
	jmp	$f4a7		; normal load
ndl_skp1:
	pla
	tay
	lda	$ba
	sta	$b8
	ldy	$b7
	bne	ndl_skp2
	jmp	$f710	;Missing filename!
ndl_skp2:
	ldx	$b9
	jsr	$f5af	;Print "SEARCHING FOR"
	lda	#$60
	sta	$b9
	jsr	$f3d5	;Iec-OPEN
	lda	$ba
	jsr	$ed09	;TALK
	lda	$b9
	jsr	$edc7	;TKSA

	jsr	$ee13	;GetByte
	sta	$ae
	lda	$90
	lsr
	lsr
	bcc	ndl_skp3
	jmp	$f704	;File Not Found
ndl_skp3:
	jsr	$ee13
	sta	$af

	txa
	bne	ndl_skp4

	lda	$c3
	sta	$ae
	lda	$c4
	sta	$af
ndl_skp4:
	ldx	#0
	jsr	ndl_PrintLoadingTo
ndl_lp1:
	lda	#$fd
	and	$90
	sta	$90
	jsr	$ffe1	;Break:
	bne	ndl_skp5

	jsr	$f642	;Iec-Close
	jsr	ndl_BreakCleanup
	rts

ndl_skp5:
	jsr	$ee13	;GetByte
	tax
	lda	$90
	lsr
	lsr
	bcs	ndl_lp1	;Reread
	txa
	ldy	#0
	jsr	PutByte_ae

	inc	$ae
	bne	ndl_skp6	;ka dest
	inc	$af
ndl_skp6:
	bit	$90
	bvc	ndl_lp1	;EOF?

	jsr	ndl_NormalCleanup
	jsr	$f642	;IEC-Close
	bcc	ndl_ex1
	jmp	$f704	;File Not Found
ndl_ex1:
	clc
	ldx	$ae
	ldy	$af
	rts
;***
ndl_BreakCleanup:
	ldx	#LoadingFromTO2_MSG-LoadingFromTO_MSG
	jsr	ndl_PrintLoadingTo
	lda	#$00
	sec
	rts
ndl_NormalCleanup:
	jsr	$edef	;Untalk
	ldx	#LoadingFromTO2_MSG-LoadingFromTO_MSG
	jsr	ndl_PrintLoadingTo
	rts
;***
LoadingFromTO_MSG:
	dc.b	13,"LOADING FROM $",0
LoadingFromTO2_MSG:
	dc.b	" TO $",0
;***
ndl_PrintLoadingTo:
	bit	$9d
	bpl	ndpl_ex1
ndlp_lp1:
	lda	LoadingFromTO_MSG,x
	beq	ndlp_skp1
	jsr	$ffd2
	inx
	bne	ndlp_lp1
ndlp_skp1:
	lda	$af
	jsr	PrintHexByte
	lda	$ae
	jsr	PrintHexByte
;***
ndpl_ex1:
	clc
	rts
;**************************************************************************


;**************************************************************************
; $c1/$c2=load address
; $ac/$ad=start address
; $ae/$af=end address+1
NormalDiskSave:
	lda	#$61
	sta	$b9
	ldy	$b7
	bne	nds_skp1
	jmp	$f710
nds_skp1:
	jsr	$f3d5	;Iec-Open
	jsr	$f68f	;print "SAVING"
	lda	$ba
	jsr	$ed0c	;LISTEN
	lda	$b9
	jsr	$edb9	;LSTSEK

	lda	$c1
	jsr	$eddd	;Write start address
	lda	$c2
	jsr	$eddd
nds_lp1:
	lda	$ac	;all done?
	cmp	$ae
	bne	nds_skp2
	lda	$ad
	cmp	$af
	beq	nds_ex1	;yes, close up
nds_skp2:
	ldy	#0
	jsr	GetByte_ac

	jsr	$eddd
	jsr	$fcdb	;increase pointer
	jsr	$ffe1	;Break
	bne	nds_lp1
nds_ex2:
	jsr	$f642	;Iec-Close
	lda	#$00
	sec
	rts
nds_ex1:
	jsr	$edfe	;UNLISTEN
	jmp	nds_ex2
;**************************************************************************

go_ind_retstore:
	jsr	install_jmp
	lda	RetStore
	sta	jmp_trmp_sm1+1
	lda	RetStore+1
	sta	jmp_trmp_sm1+2
	bra	clone_jmp
clone_e394:
	jsr	install_jmp
	lda	#$94
	sta	jmp_trmp_sm1+1
	lda	#$e3
	sta	jmp_trmp_sm1+2
	bra	clone_jmp
clone_e37b:
	jsr	install_jmp
	lda	#$7b
	sta	jmp_trmp_sm1+1
	lda	#$e3
	sta	jmp_trmp_sm1+2
clone_jmp:
	sir	$d8		; Y mapped on reg D, and X mapped on reg 8
	jmp	jmp_trmp_run
	
install_jmp:
	ldx	#jmp_trmp_end-jmp_trmp-1
ij_lp1:
	lda	jmp_trmp,x
	sta	jmp_trmp_run,x
	dex
	bpl	ij_lp1
	rts
jmp_trmp:
	rorg	$0110
jmp_trmp_run:
	ldy	#$01		;bank 1 ($004000)
	ldx	#%01010101	;all banks in RAM.
	sir	$12		;restore default x and y mapping.
jmp_trmp_sm1:
	jmp	jmp_trmp_sm1
	rend
jmp_trmp_end:


clone_ab3f	equ	$ab3f
;	jmp	$ab3f
		
clone_bdcd	equ	$bdcd
;	jmp	$bdcd
	
	
;**************************************************************************
	
GetByte_ac:
	sei
	sty	$a5
	tya
	clc
	adc	$ac
	sta	$a3
	lda	$ad
	bra	gb_common

GetByte_c1:
	sei
	sty	$a5
	tya
	clc
	adc	$c1
	sta	$a3
	lda	$c2
	bra	gb_common

GetByte_c3:
	sei
	sty	$a5
	tya
	clc
	adc	$c3
	sta	$a3
	lda	$c4
gb_common:	
	adc	#0
	tay
	and	#%00111111
	ora	#$80
	sta	$a4
	tya
	lsr
	lsr
	lsr
	lsr
	lsr
	lsr
	ora	BankStore
	sir	$e8
	tay
	ldx	RomStore
	sir	$12
	lda	MemoryStore
	sta	$01
	ldy	#0
	lda	($a3),y
	pha
	lda	#$07
	sta	$01
	sir	$e8
	ldy	#$02
	ldx	#DEFAULT_REG8
	sir	$12
	pla
	ldy	$a5
	cli
	rts
		
	
;**************************************************************************

PutByte_ae:
	sei
	pha
	sty	$a5
	tya
	clc
	adc	$ae
	sta	$a3
	lda	$af
	bra	pb_common

PutByte_c1:
	sei
	pha
	sty	$a5
	tya
	clc
	adc	$c1
	sta	$a3
	lda	$c2
	bra	pb_common

PutByte_c3:
	sei
	pha
	sty	$a5
	tya
	clc
	adc	$c3
	sta	$a3
	lda	$c4
pb_common:	
	adc	#0
	tay
	and	#%00111111
	ora	#$80
	sta	$a4
	tya
	lsr
	lsr
	lsr
	lsr
	lsr
	lsr
	ora	BankStore
	sir	$e2
	tay
	sir	$12
	lda	MemoryStore
	sta	$01
	ldy	#0
	pla
	sta	($a3),y
	lda	#$07
	sta	$01
	sir	$e2
	ldy	#$02
	sir	$12
	ldy	$a5
	cli
	rts

;**************************************************************************

;**************************************************************************
;*
;* NAME  read_d03f
;*
;* DESCRIPTION
;*   Check the contents of $d03f by checking the alias at $d040.
;*   
;******
read_d03f:
	php
	sei
	lda	$d000
	pha
	lda	#0
	sta	$39
	jsr	c_alias
	ldx	#0
	stx	$d03f
	jsr	c_alias
	inx
	stx	$d03f
	jsr	c_alias

; alias: 2=initial, 1=$d03f=0, 0=$d03f=1
	lda	$39
	lsr
	lsr			; ~$d03f bit 0
	sta	$3a
	lda	$39
	asl
	eor	$39
	and	#%00000010	; ~$d03f bit 1
	ora	$3a
	eor	#%00000011	; =$d03f

; $d03f determined, restore the value
	sta	$d03f

	tay
	pla
	sta	$d000
	plp
	tya
	rts
	
c_alias:
	lda	#$55
ca_lp1:
	sta	$d000
	cmp	$d040
	bne	ca_fl1
	eor	#$ff
	cmp	#$55
	bne	ca_lp1
; C=1
ca_ex1:
	rol	$39
	rts
ca_fl1:
	clc
	bcc	ca_ex1

	
; eof
